#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for general_format_functions.py
"""

import os
import logging
import shutil
from io import StringIO
import pytest

import PanACoTA.utils as utils
import PanACoTA.annotate_module.general_format_functions as ffunc
import test.test_unit.utilities_for_tests as tutil


ANNOTEDIR = os.path.join("test", "data", "annotate")
EXP_ANNOTE = os.path.join(ANNOTEDIR, "exp_files")
GENEPATH = os.path.join(ANNOTEDIR, "generated_by_unit-tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    # utils.init_logger(LOGFILE_BASE, 0, 'test_fastme', verbose=1)
    if os.path.isdir(GENEPATH):
        content = os.listdir(GENEPATH)
        for f in content:
            assert f.startswith(".fuse")
    else:
        os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH)
    print("teardown")


# Define variables and functions used by several tests
def my_logger():
    """
    logger given to function called by a subprocess
    """
    import multiprocessing
    m = multiprocessing.Manager()
    q = m.Queue()
    qh = logging.handlers.QueueHandler(q)
    root = logging.getLogger('test_gene-format')
    root.setLevel(logging.DEBUG)
    root.handlers = []
    logging.addLevelName(utils.detail_lvl(), "DETAIL")
    root.addHandler(qh)
    return q, logging.getLogger('process')


# Start tests
def test_write_gene():
    """
    Test that lstinfo line is written as expected when writing info for
    a gene (CDS).
    """
    gtype = "CDS"
    locus_num = "5621221"
    gene_name = "abc"
    product = "new product"
    cont_loc = "i"
    genome = "ESCO.0216.00005"
    cont_num = 15
    ecnum = "454.12.5"
    inf2 = "more information... dfd | with | pipe|characters..."
    db_xref = "mydb|pipe"
    strand = "C"
    start = str(154)
    end = str(656)
    lstfile = os.path.join(GENEPATH, "toto.lst")
    lstopenfile = open(lstfile, "w")
    lst_line = ffunc.write_gene(gtype, locus_num, gene_name, product,
                                cont_loc, genome, cont_num, ecnum, inf2, db_xref, strand,
                                start, end, lstopenfile)
    lstopenfile.close()
    assert lst_line == ("154\t656\tC\tCDS\tESCO.0216.00005.0015i_5621221\tabc\t| new product "
                        "| 454.12.5 | more information... dfd _ with _ pipe_characters... | "
                        "mydb_pipe")
    exp_file = os.path.join(EXP_ANNOTE, "res_test_write_geneCDS.lst")
    assert tutil.compare_order_content(exp_file, lstfile)


def test_write_header_gene():
    """
    From a given line of lstinfo file, giving info for a gene (start, end, gene name,
    product, EC number, more information), check that the header line of the protein and
    gene files are generated as expected.
    """
    outfile = StringIO()
    lstline = ("4416\t6068\tD\tCDS\ttest.0417.00002.0001i_00005\tyiaD\t| "
               "putative lipoprotein YiaD | 6.3.2.- | similar to AA sequence:UniProtKB:P37665")
    ffunc.write_header(lstline, outfile)
    res = outfile.getvalue()
    outfile.close()
    exp = (">test.0417.00002.0001i_00005 1653 yiaD | putative lipoprotein YiaD | 6.3.2.- "
           "| similar to AA sequence:UniProtKB:P37665\n")
    assert res == exp


def test_write_header_gene_no_name():
    """
    From a given line of lstinfo file, giving info for a gene with many unknown parts (gene
    name, product, EC number and more information are NAs), check that the header line of the
    protein and gene files are generated as expected.
    """
    outfile = StringIO()
    lstline = ("4632\t5000\tC\tCDS\ttest.0417.00002.0002b_00011\tNA\t| hypothetical protein "
               "| NA | NA")
    ffunc.write_header(lstline, outfile)
    res = outfile.getvalue()
    exp = ">test.0417.00002.0002b_00011 369 NA | hypothetical protein | NA | NA\n"
    assert res == exp
    outfile.close()


def test_handle_genome_badprok():
    """
    Test that when we try to format a genome which was annotated by prokka, but original genome
    is empty -> cannot format genome -> returns False associated with genome name
    """
    # Create prokka output dir and pur expected files (empty, we want to generate an error)
    gpath = os.path.join(GENEPATH, "toto.fasta")
    open(gpath, "w").close()
    prok_path = gpath + "-prokkaRes"
    os.makedirs(prok_path)
    fna_res = os.path.join(prok_path, " toto.fna")
    tbl_res = os.path.join(prok_path, " toto.tbl")
    gff_res = os.path.join(prok_path, "toto.gff")
    ffn_res = os.path.join(prok_path, "toto.ffn")
    faa_res = os.path.join(prok_path, "toto.faa")
    for file in [fna_res, tbl_res, gff_res, ffn_res, faa_res]:
        open(file, "w").close()
    # Create output directory for .fna file
    rep_dir = os.path.join(GENEPATH, "Replicons")
    os.makedirs(rep_dir)
    # Get args for function
    args = ("toto.fasta", "name", gpath, GENEPATH, "lst/dir", "prot/dir",
            "gene/dir", rep_dir, "gff/dir", False, my_logger()[0])
    ok_format, genome = ffunc.handle_genome(args)
    assert ok_format == False
    assert genome == "toto.fasta"


def test_handle_genome_badprodigal():
    """
    Test that when we try to format a genome which was annotated by prokka, but original genome
    is empty -> cannot format genome -> returns False associated with genome name
    """
    # Create prokka output dir and pur expected files (empty, we want to generate an error)
    gpath = os.path.join(GENEPATH, "wrong.fasta")
    open(gpath, "w").close()
    prodi_path = gpath + "-prodigalRes"
    os.makedirs(prodi_path)
    gff_res = os.path.join(prodi_path, "toto.gff")
    ffn_res = os.path.join(prodi_path, "toto.ffn")
    faa_res = os.path.join(prodi_path, "toto.faa")
    for file in [gff_res, ffn_res, faa_res]:
        open(file, "w").close()
    # Create output directory for .fna file
    rep_dir = os.path.join(GENEPATH, "Replicons")
    os.makedirs(rep_dir)
    # Get args for function
    args = ("wrong.fasta", "name", gpath, GENEPATH, "lst/dir", "prot/dir",
            "gene/dir", rep_dir, "gff/dir", True, my_logger()[0])
    ok_format, genome = ffunc.handle_genome(args)
    assert ok_format == False
    assert genome == "wrong.fasta"


def test_handle_genome_formatok(caplog):
    """
    Test that when we try to format a genome which was annotated by prokka without any problem
    It returns True associated with the genome name
    """
    caplog.set_level(logging.DEBUG)
    name = "test.0417.00002"
    # path to original genome, given to prodigal for annotation
    gpath =  os.path.join(ANNOTEDIR, "test_files", "original_name.fna")
    prok_path = os.path.join(ANNOTEDIR, "test_files")
    # Create result directories
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gene_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(prot_dir)
    os.makedirs(lst_dir)
    os.makedirs(rep_dir)
    os.makedirs(gene_dir)
    os.makedirs(gff_dir)
    # Get args for function
    args = ("original_name", name, gpath, prok_path, lst_dir, prot_dir,
            gene_dir, rep_dir, gff_dir, False, my_logger()[0])
    ok_format, genome = ffunc.handle_genome(args)
    assert ok_format == True
    assert genome == "original_name"
    # Check generated files
    exp_rep = os.path.join(EXP_ANNOTE, "res_created_rep-prokka.fna")
    res_rep_file = os.path.join(rep_dir, "test.0417.00002.fna")
    assert tutil.compare_order_content(exp_rep, res_rep_file)
    # Proteins
    exp_prt = os.path.join(EXP_ANNOTE, "res_create_prt_prokka.faa")
    res_prt_file = os.path.join(prot_dir, "test.0417.00002.prt")
    assert tutil.compare_order_content(exp_prt, res_prt_file)
    # Genes
    exp_gen = os.path.join(EXP_ANNOTE, "res_create_gene_prokka.gen")
    res_gen_file = os.path.join(gene_dir, "test.0417.00002.gen")
    assert tutil.compare_order_content(exp_gen, res_gen_file)
    # LSTINFO
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_lst-prokka.lst")
    res_lst_file = os.path.join(lst_dir, "test.0417.00002.lst")
    assert tutil.compare_order_content(exp_lst, res_lst_file)
    # gff
    exp_gff = os.path.join(EXP_ANNOTE, "res_create_gff-prokka.gff")
    res_gff_file = os.path.join(gff_dir, "test.0417.00002.gff")
    assert tutil.compare_order_content(exp_gff, res_gff_file)


def test_handle_genome_formatok_prodigal(caplog):
    """
    Test that when we try to format a genome which was annotated by prodigal without any problem
    It returns True associated with the genome name
    """
    caplog.set_level(logging.DEBUG)
    name_orig = "prodigal.outtest.ok"
    name = "test.0417.00002"
    # path to original genome, given to prodigal for annotation
    gpath =  os.path.join(ANNOTEDIR, "test_files", "original_name.fna")
    prodi_path = os.path.join(ANNOTEDIR, "test_files")
    # Create result directories
    prot_dir = os.path.join(GENEPATH, "Proteins")
    lst_dir = os.path.join(GENEPATH, "LSTINFO")
    rep_dir = os.path.join(GENEPATH, "Replicons")
    gene_dir = os.path.join(GENEPATH, "Genes")
    gff_dir = os.path.join(GENEPATH, "gff")
    os.makedirs(prot_dir)
    os.makedirs(lst_dir)
    os.makedirs(rep_dir)
    os.makedirs(gene_dir)
    os.makedirs(gff_dir)
    # Get args for function
    args = (name_orig, name, gpath, prodi_path, lst_dir, prot_dir,
            gene_dir, rep_dir, gff_dir, True, my_logger()[0])
    ok_format, genome = ffunc.handle_genome(args)
    assert ok_format == True
    assert genome == name_orig
    # Check generated files
    exp_rep = os.path.join(EXP_ANNOTE, "res_created_rep-prokka.fna")
    res_rep_file = os.path.join(rep_dir, "test.0417.00002.fna")
    assert tutil.compare_order_content(exp_rep, res_rep_file)
    # Proteins
    exp_prt = os.path.join(EXP_ANNOTE, "res_create_prt_prodigal.faa")
    res_prt_file = os.path.join(prot_dir, "test.0417.00002.prt")
    assert tutil.compare_order_content(exp_prt, res_prt_file)
    # Genes
    exp_gen = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.gen")
    res_gen_file = os.path.join(gene_dir, "test.0417.00002.gen")
    assert tutil.compare_order_content(exp_gen, res_gen_file)
    # LSTINFO
    exp_lst = os.path.join(EXP_ANNOTE, "res_create_gene_lst_prodigal.lst")
    res_lst_file = os.path.join(lst_dir, "test.0417.00002.lst")
    assert tutil.compare_order_content(exp_lst, res_lst_file)
    # gff
    exp_gff = os.path.join(EXP_ANNOTE, "res_create_gff_prodigal.gff")
    res_gff_file = os.path.join(gff_dir, "test.0417.00002.gff")
    assert tutil.compare_order_content(exp_gff, res_gff_file)


def test_format_all_prokka(caplog):
    """
    Test that when giving a list of genomes, for which prokka ran without problem,
    they are formatted, with all expected files created.
    """
    caplog.set_level(logging.DEBUG)
    # genomes = {genome: [name, gpath, to_annot, size, nbcont, l90]}
    # Get genome names we want to format (with their path)
    gnames = ["H299_H561.fasta", "B2_A3_5.fasta-changeName.fna"]
    gpaths = [os.path.join(ANNOTEDIR, "genomes", name) for name in gnames]
    onames = ["test_runprokka_H299", "test.0417.00002"]
    genomes = {gnames[0]: [onames[0], gpaths[0], gpaths[0], 12656, 3, 1],
               gnames[1]: [onames[1], gpaths[1], gpaths[1], 456464645, 5, 1]
               }
    res_path = GENEPATH
    annotated_path = os.path.join(ANNOTEDIR, "exp_files")
    # Format both genomes
    skipped_format = ffunc.format_genomes(genomes, res_path, annotated_path, False, threads=2)
    assert skipped_format == []
    # Get all names of expected output files
    exp_dir = os.path.join(EXP_ANNOTE, "res_formatAll", "prokka")
    exp_folders = ["LSTINFO", "Proteins", "Genes", "Replicons", "gff3"]
    exp_extensions = [".lst", ".prt", ".gen", ".fna", ".gff"]
    # Check that output files are created, and contain what is expected
    for fol, ext in zip(exp_folders, exp_extensions):
        exp_files = [os.path.join(exp_dir, fol, name + ext) for name in onames]
        res_files = [os.path.join(res_path, fol, name + ext) for name in onames]
        for res, exp in zip(res_files, exp_files):
            assert os.path.isfile(res)
            assert tutil.compare_order_content(res, exp)
    # Check log
    assert "Formatting all genomes" in caplog.text


def test_format_all_prodigal(caplog):
    """
    Test that when giving a list of genomes, for which prokka ran without problem,
    they are formatted, with all expected files created.
    """
    caplog.set_level(logging.DEBUG)
    # genomes = {genome: [name, gpath, to_annot, size, nbcont, l90]}
    # Get genome names we want to format (with their path)
    gnames = ["H299_H561.fasta", "B2_A3_5.fasta-changeName.fna"]
    gpaths = [os.path.join(ANNOTEDIR, "genomes", name) for name in gnames]
    onames = ["test_runprokka_H299", "test.0417.00002"]
    genomes = {gnames[0]: [onames[0], gpaths[0], gpaths[0], 12656, 3, 1],
               gnames[1]: [onames[1], gpaths[1], gpaths[1], 456464645, 5, 1]
               }
    res_path = GENEPATH
    annotated_path = os.path.join(ANNOTEDIR, "exp_files")
    # Format both genomes
    skipped_format = ffunc.format_genomes(genomes, res_path, annotated_path, True, threads=2)
    assert skipped_format == []
    # Get all names of expected output files
    exp_dir = os.path.join(EXP_ANNOTE, "res_formatAll", "prodigal")
    exp_folders = ["LSTINFO", "Proteins", "Genes", "Replicons", "gff3"]
    exp_extensions = [".lst", ".prt", ".gen", ".fna", ".gff"]
    # Check that output files are created, and contain what is expected
    for fol, ext in zip(exp_folders, exp_extensions):
        exp_files = [os.path.join(exp_dir, fol, name + ext) for name in onames]
        res_files = [os.path.join(res_path, fol, name + ext) for name in onames]
        for res, exp in zip(res_files, exp_files):
            assert os.path.isfile(res)
            assert tutil.compare_order_content(res, exp)
    # Check log
    assert "Formatting all genomes" in caplog.text


def test_format_allpb_prokka(caplog):
    """
    Test that when giving a list of genomes, 1 that is correctly formatted, and 1 has a pb,
    it returns the last one in skipped_format
    """
    caplog.set_level(logging.DEBUG)
    # Create empty original sequence files
    genomes = ["wrong.fasta", "error.fasta"]
    gpaths = [os.path.join(GENEPATH, name) for name in genomes]
    for file in gpaths:
        open(file, "w").close()
    # Add prokka (empty) result files to prokkaRes directory
    prok_paths = [gpath + "-prokkaRes" for gpath in gpaths]
    for prok_path in prok_paths:
        os.makedirs(prok_path)
        tbl_res = os.path.join(prok_path, "toto.tbl")
        gff_res = os.path.join(prok_path, "toto.gff")
        ffn_res = os.path.join(prok_path, "toto.ffn")
        faa_res = os.path.join(prok_path, "toto.faa")
        fna_res = os.path.join(prok_path, "toto.fna")
        for file in [fna_res, tbl_res, gff_res, ffn_res, faa_res]:
            open(file, "w").close()
    # Create output directory for .fna files
    rep_dir = os.path.join(GENEPATH, "Replicons")
    os.makedirs(rep_dir)
    # genomes = {genome: [name, gpath, to_annot, size, nbcont, l90]}
    genomes = {genomes[0]: ["test_wrong-fasta", gpaths[0], gpaths[0], 12656, 3, 1],
               genomes[1]: ["test_error-fasta", gpaths[1], gpaths[1], 456464645, 5, 1]
              }
    res_path = GENEPATH
    annotated_path = GENEPATH
    # Try to format both genomes
    skipped_format = ffunc.format_genomes(genomes, res_path, annotated_path, False, threads=1)
    assert skipped_format == ["wrong.fasta", "error.fasta"]
    # Get all names of expected output files
    exp_folders = ["LSTINFO", "Proteins", "Genes", "Replicons", "gff3"]
    for res_folder in [os.path.join(res_path, folder) for folder in exp_folders]:
        assert len(os.listdir(res_folder)) == 0
    # Check log
    assert "Formatting all genomes" in caplog.text
    assert ("Your genome test/data/annotate/generated_by_unit-tests/"
            "wrong.fasta-prokkaRes/toto.fna does not "
            "contain any sequence, or is not in fasta format.") in caplog.text
    assert ("Your genome test/data/annotate/generated_by_unit-tests/"
            "error.fasta-prokkaRes/toto.fna does not "
            "contain any sequence, or is not in fasta format.") in caplog.text
    assert "Problems while generating Replicon file for test_wrong-fasta" in caplog.text
    assert "Problems while generating Replicon file for test_error-fasta" in caplog.text


def test_format_1pb_prodigal(caplog):
    """
    Test that when giving a list of genomes, 1 that is correctly formatted, and 1 has a pb,
    it returns the last one in skipped_format
    """
    caplog.set_level(logging.DEBUG)
    # GENOME 2: Create empty original genome file
    genome1 = "wrong.fasta"
    gpath1 = os.path.join(GENEPATH, "wrong.fasta")
    open(gpath1, "w").close()
    # Add prodigal (empty) result files to prodigalRes directory
    prodi_path = gpath1 + "-prodigalRes"
    os.makedirs(prodi_path)
    gff_res = os.path.join(prodi_path, "toto.gff")
    ffn_res = os.path.join(prodi_path, "toto.ffn")
    faa_res = os.path.join(prodi_path, "toto.faa")
    for file in [gff_res, ffn_res, faa_res]:
        open(file, "w").close()
    # Create output directory for .fna file
    rep_dir = os.path.join(GENEPATH, "Replicons")
    os.makedirs(rep_dir)
    # GENOME 2
    genome2 = "H299_H561.fasta"
    gpath2 = os.path.join(ANNOTEDIR, "genomes", genome2)
    # Copy results of prodigal for this genome to output dir (GENEPATH)
    orig_res_files = os.path.join(EXP_ANNOTE, genome2 + '-prodigalRes')
    used_res_path = os.path.join(GENEPATH, genome2 + "-prodigalRes")
    shutil.copytree(orig_res_files, used_res_path)
    # genomes = {genome: [name, gpath, to_annot, size, nbcont, l90]}
    genomes = {genome1: ["test_genome1", gpath1, gpath1, 12656, 3, 1],
               genome2: ["test_runprokka_H299", gpath2, gpath2, 456464645, 5, 1]
               }
    res_path = GENEPATH
    annotated_path = GENEPATH
    # Format both genomes
    skipped_format = ffunc.format_genomes(genomes, res_path, annotated_path, True, threads=2)
    assert skipped_format == ["wrong.fasta"]
    # Get all names of expected output files
    exp_dir = os.path.join(EXP_ANNOTE, "res_formatAll", "prodigal")
    exp_folders = ["LSTINFO", "Proteins", "Genes", "Replicons", "gff3"]
    exp_extensions = [".lst", ".prt", ".gen", ".fna", ".gff"]
    # Check that output files are created, and contain what is expected
    for fol, ext in zip(exp_folders, exp_extensions):
        exp_file = os.path.join(exp_dir, fol, "test_runprokka_H299" + ext)
        res_file = os.path.join(res_path, fol, "test_runprokka_H299" + ext)
        assert os.path.isfile(res_file)
        assert tutil.compare_order_content(res_file, exp_file)
    # Check log
    assert "Formatting all genomes" in caplog.text
    assert ("Your genome test/data/annotate/generated_by_unit-tests/wrong.fasta does not "
            "contain any sequence, or is not in fasta format.") in caplog.text
    assert "Problems while generating Replicon file for test_genome1" in caplog.text



#!/usr/bin/env python3
# coding: utf-8

"""
Functional tests for tree subcommand
"""
import os
import subprocess
import shutil
import pytest

from PanACoTA.subcommands import tree
from test.test_unit.test_tree import utilities as tutils

# Define common variables
ALDIR = os.path.join("test", "data", "align")
TREEDIR = os.path.join("test", "data", "tree")
EXPALDIR = os.path.join(ALDIR, "exp_files")
ALIGNMENT = os.path.join(EXPALDIR, "exp_pers4genomes.grp.aln")
GENEPATH = os.path.join(TREEDIR, "generated_by_func_tests")


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    if not os.path.isdir(GENEPATH):
        os.mkdir(GENEPATH)
    print("setup")

    yield
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")


def test_main_default(capsys):
    """
    Test that when giving the alignment file, and all default parameters, it runs iqtree,
    and returns expected files
    """
    outdir = GENEPATH
    soft = "iqtree"
    model = "GTR"
    threads = 1
    verbose = 3

    cmd = "cmd test_main_default"
    tree.main(cmd, ALIGNMENT, outdir, soft, model, threads, verbose=verbose)
    # Check output files
    iq_log_file = os.path.join(outdir, "exp_pers4genomes.grp.aln.iqtree_tree.log")
    assert os.path.isfile(iq_log_file)
    with open(iq_log_file, "r") as logf:
        iq_lines = logf.readlines()
    assert ("Command: iqtree -s test/data/align/exp_files/exp_pers4genomes.grp.aln -nt 1 "
            "-m GTR -st DNA -pre test/data/tree/generated_by_func_tests/"
            "exp_pers4genomes.grp.aln.iqtree_tree -quiet") in " ".join(iq_lines)
    assert ("Alignment has 4 sequences with 6438 columns, "
            "81 distinct patterns") in " ".join(iq_lines)
    assert "Analysis results written to:" in " ".join(iq_lines)
    tree_file = os.path.join(outdir, "exp_pers4genomes.grp.aln.iqtree_tree.treefile")
    assert os.path.isfile(tree_file)
    assert tutils.is_tree_lengths(tree_file)
    assert not tutils.is_tree_bootstrap(tree_file)
    logs_base = os.path.join(outdir, "PanACoTA-tree-iqtree.log")
    assert os.path.isfile(logs_base)
    assert os.path.isfile(logs_base + ".err")
    # Check logs
    out, err = capsys.readouterr()
    print(out)
    assert "Running IQtree..." in out
    assert ("IQtree command: iqtree -s test/data/align/exp_files/exp_pers4genomes.grp.aln "
            "-nt 1 -m GTR    -st DNA -pre test/data/tree/generated_by_func_tests/"
            "exp_pers4genomes.grp.aln.iqtree_tree -quiet") in out
    assert "END" in out


def test_main_quicktree(capsys):
    """
    Test that when giving the alignment file, running with quicktree, no bootstraps,
    it creates expected files
    """
    outdir = GENEPATH
    soft = "quicktree"
    model = None
    threads = 1
    cmd = "cmd: test_main_quicktree"
    tree.main(cmd, ALIGNMENT, outdir, soft, model, threads, verbose=2)
    # Check output files
    # stockholm alignments
    stockholm = os.path.join(outdir, "exp_pers4genomes.grp.aln.stockholm")
    assert os.path.isfile(stockholm)
    # quicktree logfile
    log_file = stockholm + ".quicktree.log"
    assert os.path.isfile(log_file)
    with open(log_file, "r") as logf:
        assert logf.readlines() == []
    # tree file
    tree_file = stockholm + ".quicktree_tree.nwk"
    assert os.path.isfile(tree_file)
    assert tutils.is_tree_lengths(tree_file)
    assert not tutils.is_tree_bootstrap(tree_file)
    # log files
    logs_base = os.path.join(outdir, "PanACoTA-tree-quicktree.log")
    assert os.path.isfile(logs_base)
    assert os.path.isfile(logs_base + ".details")
    assert os.path.isfile(logs_base + ".err")
    # Check logs
    out, err = capsys.readouterr()
    assert "Converting fasta alignment to stockholm format." in out
    assert "Running Quicktree..." in out
    assert ("quicktree -in a -out t  test/data/tree/generated_by_func_tests/exp_pers4genomes.grp.aln.stockholm") in out
    assert "END" in out


def test_main_quicktree_notverbose(capsys):
    """
    Test that when giving the alignment file, running with quicktree, no bootstraps,
    it creates expected files
    """
    outdir = GENEPATH
    soft = "quicktree"
    model = None
    threads = 1
    cmd = "cmd: test_main_quicktree"
    tree.main(cmd, ALIGNMENT, outdir, soft, model, threads)
    # Check output files
    # stockholm alignments
    stockholm = os.path.join(outdir, "exp_pers4genomes.grp.aln.stockholm")
    assert os.path.isfile(stockholm)
    # quicktree logfile
    log_file = stockholm + ".quicktree.log"
    assert os.path.isfile(log_file)
    with open(log_file, "r") as logf:
        assert logf.readlines() == []
    # tree file
    tree_file = stockholm + ".quicktree_tree.nwk"
    assert os.path.isfile(tree_file)
    assert tutils.is_tree_lengths(tree_file)
    assert not tutils.is_tree_bootstrap(tree_file)
    # log files
    logs_base = os.path.join(outdir, "PanACoTA-tree-quicktree.log")
    assert os.path.isfile(logs_base)
    assert os.path.isfile(logs_base + ".details")
    assert os.path.isfile(logs_base + ".err")
    # Check logs
    out, err = capsys.readouterr()
    assert "Converting fasta alignment to stockholm format." in out
    assert "Running Quicktree..." in out
    assert "END" in out


def test_main_fastme(capsys):
    """
    Test that when giving the alignment file, running with fastme, no bootstraps,
    it creates expected files
    """
    outdir = GENEPATH
    soft = "fastme"
    model = None
    threads = None
    boot = 100
    cmd = "cmd: test_main_quicktree"
    tree.main(cmd, ALIGNMENT, outdir, soft, model, threads, boot=boot, verbose=16, write_boot=True)
    # Check output files
    # phylip alignments
    phylip = os.path.join(outdir, "exp_pers4genomes.grp.aln.phylip")
    assert os.path.isfile(phylip)
    # fastme logfile
    log_file = phylip + ".fastme.log"
    assert os.path.isfile(log_file)
    with open(log_file, "r") as logf:
        fastme_lines = logf.readlines()
    assert "Input data type  DNA" in " ".join(fastme_lines)
    assert "evolutionary model  TN93" in " ".join(fastme_lines)
    # tree file
    tree_file = phylip + ".fastme_tree.nwk"
    assert os.path.isfile(tree_file)
    assert not tutils.is_tree_lengths(tree_file)
    assert tutils.is_tree_bootstrap(tree_file)
    # bootstrap file
    tree_boot_file = phylip + ".fastme_bootstraps.nwk"
    assert os.path.isfile(tree_boot_file)
    with open(tree_boot_file, "r") as tbf:
        lines = tbf.readlines()
    assert len(lines) == boot
    # log files
    logs_base = os.path.join(outdir, "PanACoTA-tree-fastme.log")
    assert os.path.isfile(logs_base)
    assert os.path.isfile(logs_base + ".details")
    assert os.path.isfile(logs_base + ".debug")
    assert os.path.isfile(logs_base + ".err")
    # Check logs
    out, err = capsys.readouterr()
    assert "Converting fasta alignment to PHYLIP-relaxed format." in out
    assert "Running FastME..." in out
    assert ("fastme -i test/data/tree/generated_by_func_tests/exp_pers4genomes.grp.aln.phylip "
            "-dT -nB -s  -b 100 -o test/data/tree/generated_by_func_tests/"
            "exp_pers4genomes.grp.aln.phylip.fastme_tree.nwk "
            "-I test/data/tree/generated_by_func_tests/exp_pers4genomes.grp.aln.phylip.fastme.log "
            "-B test/data/tree/generated_by_func_tests/"
            "exp_pers4genomes.grp.aln.phylip.fastme_bootstraps.nwk") in out
    assert "END" in out


def test_main_fasttree(capsys):
    """
    Test that when giving the alignment file, running with fasttree, no bootstraps,
    it creates expected files
    """
    outdir = GENEPATH
    soft = "fasttree"
    model = "-gtr"
    threads = 1
    boot = 100
    cmd = "cmd: test_main_fasttree"
    tree.main(cmd, ALIGNMENT, outdir, soft, model, threads, boot=boot, verbose=2)
    # Check output files
    # fastme logfile
    log_file = os.path.join(outdir, "exp_pers4genomes.grp.aln.fasttree.log")
    assert os.path.isfile(log_file)
    with open(log_file, "r") as logf:
        fasttree_lines = logf.readlines()
    assert "Read 4 sequences, 6438 positions" in " ".join(fasttree_lines)
    assert "TreeCompleted" in " ".join(fasttree_lines)
    assert ("Nucleotide distances: Jukes-Cantor Joins: balanced Support: "
            "Local boot 100") in " ".join(fasttree_lines)
    # tree file
    tree_file = os.path.join(outdir, "exp_pers4genomes.grp.aln.fasttree_tree.nwk")
    assert os.path.isfile(tree_file)
    assert not tutils.is_tree_lengths(tree_file)
    assert tutils.is_tree_bootstrap(tree_file)
    # log files
    logs_base = os.path.join(outdir, "PanACoTA-tree-fasttree.log")
    assert os.path.isfile(logs_base)
    assert os.path.isfile(logs_base + ".details")
    assert not os.path.isfile(logs_base + ".debug")
    assert os.path.isfile(logs_base + ".err")
    # Check logs
    out, err = capsys.readouterr()
    print(out)
    assert "Running FasttreeMP..." in out
    assert ("Fasttree command: FastTreeMP -nt -gtr -noml -nocat -boot 100 "
            "-log test/data/tree/generated_by_func_tests/exp_pers4genomes.grp.aln.fasttree.log "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln") in out
    assert "END" in out


def test_main_iqtree2_newdir(capsys):
    """
    Test that when giving the alignment file, and all default parameters, it runs iqtree2,
    and returns expected files
    """
    outdir = os.path.join(GENEPATH, "test_iqtree2")
    soft = "iqtree2"
    model = "GTR"
    threads = 1

    cmd = "cmd test_main_default"
    tree.main(cmd, ALIGNMENT, outdir, soft, model, threads, boot=1000, write_boot=True,
              verbose=2)
    # Check output files
    # Check iqtree logfile
    iq_log_file = os.path.join(outdir, "exp_pers4genomes.grp.aln.iqtree_tree.log")
    assert os.path.isfile(iq_log_file)
    with open(iq_log_file, "r") as logf:
        iq_lines = logf.readlines()
    assert ("Command: iqtree2 -s test/data/align/exp_files/exp_pers4genomes.grp.aln "
            "-T 1 -m GTR -B 1000 --boot-trees --seqtype DNA "
            "--prefix test/data/tree/generated_by_func_tests/"
            "test_iqtree2/exp_pers4genomes.grp.aln.iqtree_tree "
            "--quiet") in " ".join(iq_lines)
    assert ("Alignment has 4 sequences with 6438 columns, "
            "81 distinct patterns") in " ".join(iq_lines)
    assert "Analysis results written to:" in " ".join(iq_lines)
    # Check treefile
    tree_file = os.path.join(outdir, "exp_pers4genomes.grp.aln.iqtree_tree.treefile")
    assert os.path.isfile(tree_file)
    assert not tutils.is_tree_lengths(tree_file)
    assert tutils.is_tree_bootstrap(tree_file)
    # Check bootstrap tree file
    boot_file = os.path.join(outdir, "exp_pers4genomes.grp.aln.iqtree_tree.ufboot")
    assert os.path.isfile(tree_file)
    with open(boot_file, "r") as tbf:
        lines = tbf.readlines()
    assert len(lines) == 1000
    # Check panacota logfile
    logs_base = os.path.join(outdir, "PanACoTA-tree-iqtree2.log")
    assert os.path.isfile(logs_base)
    assert os.path.isfile(logs_base + ".err")
    assert os.path.isfile(logs_base + ".details")
    # Check logs
    out, err = capsys.readouterr()
    assert "Running IQtree..." in out
    assert ("IQtree command: iqtree2 -s test/data/align/exp_files/exp_pers4genomes.grp.aln "
            "-T 1 -m GTR  -B 1000 --boot-trees --seqtype DNA --prefix test/data/tree/"
            "generated_by_func_tests/test_iqtree2/exp_pers4genomes.grp.aln.iqtree_tree "
            "--quiet") in out
    assert "END" in out


def test_main_from_parse(capsys):
    """
    Test main when we give the output of the parser
    run fastme with K2P model, no bootstrap
    """
    import argparse
    args = argparse.Namespace()
    args.alignment = ALIGNMENT
    args.boot = None
    args.outdir = GENEPATH
    args.soft = "iqtree"
    args.model = "HKY"
    args.write_boot = False
    args.threads = 1
    args.verbose = 2
    args.quiet = False
    args.memory = False
    args.fast = False
    args.argv = "PanACoTA tree test_main_from_parse"
    tree.main_from_parse(args)
    # Check output files
    # iqtree log file
    iq_log_file = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.log")
    assert os.path.isfile(iq_log_file)
    with open(iq_log_file, "r") as logf:
        iq_lines = logf.readlines()
    assert ("Command: iqtree -s test/data/align/exp_files/exp_pers4genomes.grp.aln -nt 1 "
            "-m HKY -st DNA -pre test/data/tree/generated_by_func_tests/"
            "exp_pers4genomes.grp.aln.iqtree_tree -quiet") in " ".join(iq_lines)
    assert ("Alignment has 4 sequences with 6438 columns, "
            "81 distinct patterns") in " ".join(iq_lines)
    assert "Analysis results written to:" in " ".join(iq_lines)
    # Tree file
    tree_file = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.treefile")
    assert os.path.isfile(tree_file)
    assert tutils.is_tree_lengths(tree_file)
    assert not tutils.is_tree_bootstrap(tree_file)
    # Panacota log file
    logs_base = os.path.join(GENEPATH, "PanACoTA-tree-iqtree.log")
    assert os.path.isfile(logs_base)
    assert os.path.isfile(logs_base + ".err")
    # Check logs
    out, err = capsys.readouterr()
    assert "Running IQtree..." in out
    assert ("IQtree command: iqtree -s test/data/align/exp_files/exp_pers4genomes.grp.aln "
            "-nt 1 -m HKY    -st DNA -pre test/data/tree/generated_by_func_tests/"
            "exp_pers4genomes.grp.aln.iqtree_tree -quiet") in out
    assert "END" in out



#!/usr/bin/env python3

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions helping for doing quality control on genomes in order to eliminate
bad quality sequences, and then run Mash loops in order to discard too close genomes.

@author gem
August 2019
"""

import os
import sys
import glob
import logging
import progressbar
import scipy.sparse
from scipy.sparse import dok_matrix

from PanACoTA import utils
from PanACoTA.annotate_module import genome_seq_functions as gfunc

logger = logging.getLogger("prepare.filter")


def check_quality(species_linked, db_path, tmp_dir, max_l90, max_cont, cutn):
    """
    Do a quality control of all genomes in db_path

    Parameters
    ----------
    outdir : str
        directory for all results (refseq downloads, database init etc)
    species_linked : str
        given NCBI species with '_' instead of spaces, or NCBI taxID if species
        name not given
    dbpath : str
        directory to 'Database_init' containing all .fna files
    tmp_dir : str
        directory where all tmp files must be saved (files cut at each stretch of 'x' N)
    max_l90 : int
        max L90 value tolerated to keep a genome
    max_cont : int
        Max number of contigs tolerated to keep a genome
    cutn : int
        cut at each stretch of this number of 'N'. Don't cut if equal to 0

    Returns
    -------
        genomes : {genome_file: [genome_name, orig_path, path_to_seq_to_annotate, size,
                                 nbcont, l90]}
        no need for small name, we won't annotate genomes. genome_name is the same as filename
        but without extension
    """
    # Check database folder exists
    if not os.path.isdir(db_path):
        logger.error(f"{db_path} does not exist.")
        sys.exit(1)
    if not os.path.isdir(tmp_dir):
        logger.error(f"{tmp_dir} does not exist.")
        sys.exit(1)
    # Get all genome filenames
    all_genomes = os.listdir(db_path)
    if len(all_genomes) == 0:
        logger.error(f"There is no genome in {db_path}.")
        sys.exit(1)
    # Get name of genomes without extension
    genomes = {g:[os.path.splitext(g)[0]] for g in all_genomes}
    logger.info("Total number of genomes for {}: {}".format(species_linked, len(all_genomes)))

    # cut at stretches of 'N' if asked, and get L90, nbcontig, size for all genomes
    # -> {genome_file: [genome_g, orig_path, to_annotate_path, size, nbcont, l90]}
    gfunc.analyse_all_genomes(genomes, db_path, tmp_dir, cutn, "prepare", logger, quiet=False)
    return genomes

def sort_genomes_minhash(genomes, max_l90, max_cont):
    """
    Sort genomes:
    - draft genomes, sorted by L90 and then nb_contigs

    Parameters
    ----------

    genomes : {genome_file: [genome_name, orig_name, path_to_seq_to_annotate, size,
                             nbcont, l90]}
    max_l90 : int
        max L90 value tolerated to keep a genome
    max_cont : int
        Max number of contigs tolerated to keep a genome

    Returns
    -------
    sorted_genomes: list of 'genome_file' for all genomes kept (L90 and nbcont ok),
    ordered by decreasing quality
    """
    logger.info("Sorting all {} genomes by quality".format(len(genomes)))
    sorted_genomes = []
    # Add all genomes, sorted by L90 and then nb_cont, if L90 <= 100 and nb_cont <= 999
    nbcomp = 0
    nb_disc = 0
    for ginfo in sorted(genomes.items(), key=utils.sort_genomes_l90_nbcont):
        gname, (_, _, _, _, nb_cont, L90) = ginfo
        if L90 > max_l90 or nb_cont > max_cont:
            nb_disc += 1
        else:
            sorted_genomes.append(gname)
    logger.info(f"{len(sorted_genomes)} genomes after quality control ({nb_disc} "
                 "discarded)")
    return sorted_genomes


def iterative_mash(sorted_genomes, genomes, outdir, species_linked, min_dist, max_dist,
                   threads, quiet):
    """
    Run mash all vs all, to get all pairwise distances.
    Then, take the first genome of the list, and remove those for which the distance to it
    is not between 1e-4 and 0.06. Restart with the next genome kept in the list, and so on
    until the last genome.

    Parameters
    ----------
    sorted_genomes: list
        list of 'genome_file' for all genomes kept (L90 and nbcont ok)
    genomes : dict
        {genome_file: [genome_name, orig_name, path_to_seq_to_annotate, size, nbcont, l90]}
    outdir : str
        path to directory where all results are saved
    species_linked : str
        species name if given, otherwise species taxID
    min_dist : float
        lower limit of distance between 2 genomes to keep them
    max_dist : float
        max limit of distance between 2 genomes to keep them
    threads :
        max number of threads to use
    quiet : bool
        True if nothing must be sent to stdout/stderr, False otherwise

    Returns
    -------

    genomes_removed : dict
        {genome_name: [ref_name, dist]} genome against which 'genome_name' is removed, and corresponding distance (justifying removal)
    """
    logger.info("Starting filtering steps according to distance between genomes.")
    # Run mash all vs all
    mash_dir = os.path.join(outdir, "mash_files")
    os.makedirs(mash_dir, exist_ok=True)
    # List of genome sequences to compare
    list_reps = os.path.join(mash_dir, f"list-to-sketch-{species_linked}.txt")
    # Mash logfile
    mash_log = os.path.join(mash_dir, f"mash-all-{species_linked}.log")
    # Binary file generated by minhash sketch (index of all sequences)
    out_msh = os.path.join(mash_dir, f"all-genomes-{species_linked}")
    # Matrix with pairwise distances between all genomes
    matrix = os.path.join(mash_dir, f"matrix-all-genomes-{species_linked}.txt")
    # Binary file to save matrix of pairwise distances
    sparse_mat = os.path.join(mash_dir, f"matrix-all-genomes-{species_linked}.npz")

    # Sketch genomes
    sketch_all(genomes, sorted_genomes, outdir, list_reps, out_msh, mash_log, threads)

    # Compute pairwise distances
    compare_all(out_msh, matrix, sparse_mat, mash_log, threads)
    # Iteratively discard genomes
    # List of genomes to compare to the next ones until a limit value is reached.
    # genomes ordered by decreasing L90/nbcont (used to pop elements in comparing step)
    to_try = sorted_genomes[::-1]
    # Put list of genomes removed by mash comparison, and why
    # (out of limits distance with which genome)
    genomes_removed = {}  # {genome: [compared_with, dist]}
    nbgen = len(to_try)

    # Get link between genome_file (genomes key) and place in sorted_genomes
    corresp_file = {genome: num for num, genome in enumerate(sorted_genomes)}
    # Read matrix (from npz file if existing, otherwise from txt file)
    if os.path.exists(sparse_mat):
        logger.info(f"Loading matrix contained in {sparse_mat}")
        # convert matrix returned by load_npz (coo format, as saved) to dok format
        mat_sp = scipy.sparse.load_npz(sparse_mat).todok()
    # Read matrix txt file generated by minhash, and save this python object matrix to a npz file.
    else:
        logger.info("Reading matrix from txt file generated by Mash.")
        mat_sp = read_matrix(genomes, sorted_genomes, matrix)
        # corresp_file = {genome_file : num of genome in sorted_genomes}
        logger.info("Saving matrix to npz file to be loaded quicker if needed later")
        # Convert dok_matrix to coo format, as dok format is not allowed by save_npz
        coo_mat = mat_sp.tocoo()
        scipy.sparse.save_npz(sparse_mat, coo_mat)

    # Iteratively discard genomes too close or too far
    logger.info("Starting iterative discarding steps")
    if not quiet:
        widgets = ['Genomes compared: ',
                   progressbar.Bar(marker='█', left='', right='', fill=' '), ' ',
                   progressbar.Counter(), "/{}".format(nbgen), ' ',
                   progressbar.Timer(), ' - '
                  ]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=len(to_try), term_width=79).start()
        done = 0

    while len(to_try) > 1:
        mash_step(to_try, corresp_file, mat_sp, genomes_removed, min_dist, max_dist)
        if not quiet:
            done = nbgen - len(to_try)
            bar.update(done)
    if not quiet:
        bar.finish()
    logger.info("Final number of genomes in dataset: {}".format(nbgen - len(genomes_removed)))
    return genomes_removed


def sketch_all(genomes, sorted_genomes, outdir, list_reps, out_msh, mash_log, threads):
    """
    Sketch all genomes to a combined archive.

    Parameters
    ----------
    genomes : dict
        {genome_file: [genome_name, orig_name, path_to_seq_to_annotate, size, nbcont, l90]}
    sorted_genomes: list
        list of 'genome_file' for all genomes kept (L90 and nbcont ok), ordered by
        decreasing quality
    outdir : str
        path to directory where all results are saved
    list_reps : str
        file with list of genomes to sketch. File will be emptied if it contain something, and
        filled with the informations from 'genomes'.
    out_msh : str
        output of mash
    mash_log : str
        mash logfile
    threads :
        max number of threads to use

    Returns
    -------

    return value (0 if OK, 1 if error)

    """
    # If given outdir does not exist, close it
    if not os.path.isdir(outdir):
        logger.error(f"Your output directory '{outdir}' does not exist.")
        sys.exit(1)
    # Empty list_reps file
    open(list_reps, "w").close()
    # Complete paths to genomes to compare: 'path_to_seq_to_annotate' = genome_file[2]
    file_paths = [genomes[g][2] for g in sorted_genomes]
    # Write list of genomes to compare to a file
    utils.write_list(file_paths, list_reps)
    # Sketch all genome sequences if not already done
    if os.path.isfile(out_msh + ".msh"):
        logger.warning(f"Mash sketch file {out_msh}.msh already exists. PanACoTA will "
                        "use it for next step.")
        os.remove(list_reps)
        return 0
    logger.info("Sketching all genomes...")
    cmd_sketch = f"mash sketch -o {out_msh} -p {threads} -l {list_reps} -s 1e4"
    logger.details(cmd_sketch)
    error_sketch = (f"Error while trying to sketch {len(sorted_genomes)} genomes to combined "
                    "archive. Maybe some genome sequences in "
                    "'tmp_files' are missing! Check logfile: "
                    f"{mash_log}")

    outf = open(mash_log, "w")
    utils.run_cmd(cmd_sketch, error_sketch, eof=True, stdout=outf, stderr=outf, logger=logger)
    outf.close()
    return 0


def compare_all(out_msh, matrix, npz_matrix, mash_log, threads):
    """
    Comparing all pairwise genomes that are already been sketched in the given file.

    Parameters
    ----------
    out_msh : str
        output of mash
    matrix : str
        File to put generated matrix of pairwise distances between all genomes
    npz_matrix : str
        matrix of pairwise distances saved in a binary file
    mash_log : str
        mash logfile
    threads :
        max number of threads to use

    Returns
    -------

    return code
    """
    # txt matrix already exists
    if os.path.isfile(matrix):
        logger.warning("Matrix file {} already exists. The program will use this distance matrix "
                       "to filter all genomes according to their distances.".format(matrix))
        return 0
    # npz matrix already exists
    if os.path.isfile(npz_matrix):
        logger.warning("Matrix file {} already exists. The program will use this distance matrix "
                       "to filter all genomes according to their distances.".format(matrix))
        return 0
    logger.info("Computing pairwise distances between all genomes")
    cmd_dist = f"mash dist -p {threads} {out_msh}.msh {out_msh}.msh"
    logger.details(cmd_dist)
    # Open matfile to write matrix inside
    matfile = open(matrix, "w")
    # Open mash log to add log of 'mash dist' to log of 'mash sketch'
    outf = open(mash_log, "a")
    error_dist = ("Error while trying to estimate pairwise distances between all genomes. "
                  f"See {mash_log}.")
    utils.run_cmd(cmd_dist, error_dist, eof=True, stdout=matfile, stderr=outf)
    outf.close()
    matfile.close()
    return 0


def mash_step(to_try, corresp, mat_sp, genomes_removed, min_dist, max_dist):
    """
    Prepare a mash run, with a given genome as reference, and others to compare to.

    Parameters
    ----------
    to_try : list
        list of genome_file (keys of 'genomes') to compare, ordered by decreasing L90/nbcont
    corresp : dict
        {genome_file : num_of_genome in sorted_genomes}
    mat_sp : scipy.sparse.dok.dok_matrix
        triangle matrix containing pairwise distance comparisons
    genomes_removed : dict
        {genome_file: [ref_name, dist]} genome against which 'genome_name' is removed, and
        corresponding distance (justifying removal)
    min_dist : float
        lower limit of distance between 2 genomes to keep them
    max_dist : float
        max limit of distance between 2 genomes to keep them

    Returns
    -------

    to_try is updated (reference element and all genomes not compatible with it are removed)
    genomes_removed is updated
    return code (0 if no problem)

    """
    # Get last element (which is the 'best' genome), and remove it from the list
    ref_name = to_try.pop()
    # Line of genome in mat_sp
    ref_num = corresp[ref_name]
    # Genomes (ordered by increasing L90/nbcont) to compare to the selected element (ref_name)
    others = to_try[::-1]

    # For each genome, compare its distance to reference genome 'ref_name'
    for gname in others:
        # Column of genome in mat_sp
        other_num = corresp[gname]
        # Get distance between the 2 genomes
        if ref_num < other_num:
            dist = mat_sp[ref_num, other_num]
        else:
            logger.warning("Should never happen as mat_sp is a triangle matrix!")
            dist = mat_sp[other_num, ref_num]
        # If distance not in the limits, remove genome from to_try and add to genomes_removed list
        if not min_dist <= dist <= max_dist:
            to_try.remove(gname)
            genomes_removed[gname] = [ref_name, dist]
    return 0


def read_matrix(genomes, sorted_genomes, matrix):
    """
    Read the matrix of pairwise distances between all genomes, and save it to a sparse
    matrix (only upper triangle).

    Parameters
    ----------
    genomes : dict
        {genome_file: [genome_name, orig_name, path_to_seq_to_annotate, size, nbcont, l90]}
    sorted_genomes: list
        list of 'genome_file' for all genomes kept (L90 and nbcont ok)
    matrix : str
        File containing the matrix of pairwise distances between all genomes

    Returns
    -------

    mat_sp : str
        python dok_matrix object
    """
    if not os.path.isfile(matrix):
        logger.error(f"Matrix file {matrix} does not exist. We cannot read it "
                     "and do the next steps. Program ending.")
        sys.exit(1)

    nbgen = len(sorted_genomes)
    corresp_abs = {genomes[genome][2]: num for num, genome in enumerate(sorted_genomes)}
    # Create square matrix with nbgen cols/lines. dok format is a 'Dictionary Of Keys'
    # -> writes (0, 1) value
    mat_sp = dok_matrix((nbgen, nbgen), dtype=float)
    # Write matrix values
    with open(matrix, "r") as matf:
        for line in matf:
            path1, path2, dist = line.split()[:3]
            num1 = corresp_abs[path1]
            num2 = corresp_abs[path2]
            # only in lower triangle (no duplicate)
            if num1 < num2:
                mat_sp[num1, num2] = float(dist)
            else:
                mat_sp[num2, num1] = float(dist)
    return mat_sp


def write_outputfiles(genomes, sorted_genomes, genomes_removed, outdir, gspecies, min_dist, max_dist):
    """
    Write the list of genomes kept in a file, 1 genome per line -> will be the input file for
    annotation and next steps
    Write discarded genomes to another file, with, for each line:
    - genome name
    - problem when compared with which other genome
    - distance to this other genomes

    Parameters
    ----------
    genomes : dict
        {genome_file: [genome_name, orig_name, path_to_seq_to_annotate, size, nbcont, l90]}
    sorted_genomes: list
        list of 'genome_file' for all genomes kept (L90 and nbcont ok)
    genomes_removed : dict
        {genome_name: [ref_name, dist]} genome against which 'genome_name' is removed, and corresponding distance (justifying removal)
    outdir : str
        directory where those list files must be created
    gspecies : str
        species name if given, otherwise species taxID
    min_dist : float
        lower limit of distance between 2 genomes to keep them
    max_dist : float
        upper limit of distance between 2 genomes to keep them

    Returns
    -------
    return code
    """
    if not os.path.isdir(outdir):
        logger.error(f"The given output directory ({outdir}) does not exist. We cannot "
                      "create output files there")
        sys.exit(1)
    list_file = os.path.join(outdir, f"LSTINFO-{gspecies}-filtered-{min_dist}_{max_dist}.txt")
    kept_genomes = []
    discard_file = os.path.join(outdir, f"discarded-by-minhash-{gspecies}-{min_dist}_{max_dist}.txt")

    # Get list of kept genomes and write them in list_file
    for genome in sorted_genomes:
        if genome not in genomes_removed:
            kept_genomes.append(genome)

    # Write 4 columns in LSTINFO file (genomes kept) :
    # path to file analyzed, genome size, nbcont and L90
    with open(list_file, "w") as lf:
        lf.write('to_annotate\tgsize\tnb_conts\tL90\n')
        # For each genome in kept_genomes, find required information on this it, using 'genomes'
        for g in kept_genomes:
            _, _, analyzed, size, nbcont, l90 = genomes[g]
            towrite = utils.list_to_str([analyzed, size, nbcont, l90], sep="\t")
            lf.write(towrite)

    # Write list of discarded genomes and why they are discarded
    with open(discard_file, "w") as disf:
        disf.write("genome_name\tproblem_compared_with\tdist\n")
        for genome, info in genomes_removed.items():
            disf.write(utils.list_to_str([genome] + info))
    logger.info(f"Final list of genomes in the dataset: {list_file}")
    logger.info(f"List of genomes discarded by minhash steps: {discard_file}")
    return list_file



#!/usr/bin/env python3
# coding: utf-8

# ###############################################################################
# This file is part of PanACOTA.                                                #
#                                                                               #
# Authors: Amandine Perrin                                                      #
# Copyright © 2018-2020 Institut Pasteur (Paris).                               #
# See the COPYRIGHT file for details.                                           #
#                                                                               #
# PanACOTA is a software providing tools for large scale bacterial comparative  #
# genomics. From a set of complete and/or draft genomes, you can:               #
#    -  Do a quality control of your strains, to eliminate poor quality         #
# genomes, which would not give any information for the comparative study       #
#    -  Uniformly annotate all genomes                                          #
#    -  Do a Pan-genome                                                         #
#    -  Do a Core or Persistent genome                                          #
#    -  Align all Core/Persistent families                                      #
#    -  Infer a phylogenetic tree from the Core/Persistent families             #
#                                                                               #
# PanACOTA is free software: you can redistribute it and/or modify it under the #
# terms of the Affero GNU General Public License as published by the Free       #
# Software Foundation, either version 3 of the License, or (at your option)     #
# any later version.                                                            #
#                                                                               #
# PanACOTA is distributed in the hope that it will be useful, but WITHOUT ANY   #
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS     #
# FOR A PARTICULAR PURPOSE. See the Affero GNU General Public License           #
# for more details.                                                             #
#                                                                               #
# You should have received a copy of the Affero GNU General Public License      #
# along with PanACOTA (COPYING file).                                           #
# If not, see <https://www.gnu.org/licenses/>.                                  #
# ###############################################################################

"""
Functions to convert prokka or prodigal results to gembase format:

    * Proteins: multifasta with all CDS in aa
    * Replicons: multifasta of genome
    * Genes: multifasta of all genes in nuc
    * gff3: gff files without sequence
    * LSTINFO:
        - if annotated by prokka: information on annotation. Columns are:
        "start end strand type locus gene_name | product | EC_number | inference2"
        with the same types as prokka file, and strain is C (complement) or D (direct).
        Locus is: `<genome_name>.<contig_num><i or b>_<protein_num>`
        - if annotated by prodigal

@author gem
May 2019
"""

import os
import logging
import logging.handlers
import progressbar
import multiprocessing
import threading
import PanACoTA.utils as utils
from PanACoTA.annotate_module import format_prokka as fprokka
from PanACoTA.annotate_module import format_prodigal as fprodigal


main_logger = logging.getLogger("annotate.geneffunc")


def format_genomes(genomes_ok, res_path, annot_path, prodigal_only, threads=1, quiet=False):
    """
    For all genomes which were annotated (by prokka or prodigal), reformat them
    in order to have, in 'res_path', the following folders:

    * LSTINFO: containing a .lst file for each genome, with all genes
    * Replicons: containing all multifasta files
    * Genes: containing 1 multi-fasta per genome, with all its genes in nuc
    * Proteins: containing 1 multi-fasta per genome, with all its proteins in aa
    * gff: containing all gff files

    Parameters
    ----------
    genomes_ok : dict
        genomes to format (annotation was OK) ->
        {genome: [name, gpath, to_annot, size, nbcont, l90]}
    res_path : str
        path to folder where the 4 directories must be created
    annot_path : str
        path to folder containing "<genome_name>-[prokka, prodigal]Res" where all prokka/prodigal
        results are saved.
    prodigal_only: True if it was annotated by prodigal, False if annotated by prokka
    threads : int
        number of threads to use to while formatting genomes
    quiet : bool
        True if nothing must be sent to stderr/stdout, False otherwise

    Returns
    -------
    skipped_format : list

        list of genomes skipped because they had a problem in format step
    """
    main_logger.info("Formatting all genomes")
    lst_dir = os.path.join(res_path, "LSTINFO")
    prot_dir = os.path.join(res_path, "Proteins")
    gene_dir = os.path.join(res_path, "Genes")
    rep_dir = os.path.join(res_path, "Replicons")
    gff_dir = os.path.join(res_path, "gff3")
    os.makedirs(lst_dir, exist_ok=True)
    os.makedirs(prot_dir, exist_ok=True)
    os.makedirs(gene_dir, exist_ok=True)
    os.makedirs(rep_dir, exist_ok=True)
    os.makedirs(gff_dir, exist_ok=True)

    # If this function goes until here, it means that there is at least 1 genome to annotate
    nbgen = len(genomes_ok)
    bar = None
    if not quiet:
        # Create progressbar
        widgets = ['Formatting genomes: ',
                   progressbar.Bar(marker='█', left='', right=''),
                   ' ', progressbar.Counter(), "/{}".format(nbgen), ' (',
                   progressbar.Percentage(), ") - ", progressbar.Timer()]
        bar = progressbar.ProgressBar(widgets=widgets, max_value=nbgen, term_width=79).start()
    # Create a Queue to put logs from processes, and handle them after from a single thread
    m = multiprocessing.Manager()
    q = m.Queue()

    # if at least 1 genome ok, try to format it
    # arguments for 'handle_genome' function:
    # (genome, name, gpath, annot_path, lst_dir, prot_dir, gene_dir, rep_dir,
    # gff_dir, results, prodigal_only, q)
    params = [(genome, name, gpath, annot_path, lst_dir, prot_dir, gene_dir,
               rep_dir, gff_dir, prodigal_only, q)
              for genome, (name, _, gpath, _, _, _) in genomes_ok.items()]

    # Create pool and launch parallel formating steps
    pool = multiprocessing.Pool(threads)
    final = pool.map_async(handle_genome, params, chunksize=1)
    pool.close()
    lp = threading.Thread(target=utils.logger_thread, args=(q,))
    lp.start()
    if not quiet:
        while True:
            if final.ready():
                break
            remaining = final._number_left
            bar.update(nbgen - remaining)
        bar.finish()
    pool.join()
    q.put(None)
    lp.join()
    res = final.get()

    # List of genomes for which format step had problems: will be filled after
    skipped_format = []
    for output in res:
        if not output[0]:
            skipped_format.append(output[1])
    return skipped_format


def handle_genome(args):
    """
    For a given genome, check if it has been annotated (in results), if annotation
    (by prokka or prodigal) ran without problems (result = True). In that case,
    format the genome and get the output to see if everything went ok.

    Parameters
    ----------
    args : tuple
        (genome, name, gpath, prok_path, lst_dir, prot_dir,\
         gene_dir, rep_dir, gff_dir, results, q)\
         with:

         * genome : original genome name
         * name : gembase name of the genome
         * gpath : path to the genome sequence which was given to prokka/prodigal for annotation
         * annot_path : directory where prokka/prodigal folders are saved
         * lst_dir : path to 'LSTINFO' folder
         * prot_dir : path to 'Proteins' folder
         * gene_dit : path to 'Genes' folder
         * rep_dir : path to 'Replicons' folder
         * gff_dir : path to 'gff3' folder
         * prodigal_only : True if annotated by prodigal, False if annotated by prokka
         * q : multiprocessing.managers.AutoProxy[Queue] queue to put logs during subprocess

    Returns
    -------
    (bool, str) :

        * True if genome was annotated as expected, False otherwise
        * genome name (used to get info from the pool.map_async)
    """
    (genome, name, gpath, annot_path, lst_dir, prot_dir,
     gene_dir, rep_dir, gff_dir, prodigal_only, q) = args

    # Define which formatting must be used, given the annotation software
    if prodigal_only:
        format_one_genome = fprodigal.format_one_genome
    else:
        format_one_genome = fprokka.format_one_genome
    # Set logger for this process
    qh = logging.handlers.QueueHandler(q)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers = []
    logging.addLevelName(utils.detail_lvl(), "DETAIL")
    root.addHandler(qh)
    logger = logging.getLogger('format.handle_genome')
    # Handle genome
    ok_format = format_one_genome(gpath, name, annot_path, lst_dir,
                                  prot_dir, gene_dir, rep_dir, gff_dir)
    return ok_format, genome


def write_gene(gtype, locus_num, gene_name, product, cont_loc,
               genome, cont_num, ecnum, inf2, db_xref, strand, start, end, lstopenfile):
    """
    Write given gene to output file

    Parameters
    ----------
    gtype : str
        type of feature (CDS, tRNA, etc.)
    locus_num : str
        number of the locus given by prokka/prodigal
    gene_name : str
        gene name found by prokka/prodigal ("NA" if no gene name -> Always the case with Prodigal)
    product : str
        found by prokka/Prodigal, "NA" if no product (always the case for prodigal)
    cont_loc : str
        'i' if the gene is inside a contig, 'b' if its on the border (first or last gene
        of the contig)
    genome : str
        genome name (spegenus.date.strain_num)
    cont_num : int
        contig number
    ecnum : str
        EC number, found by prokka, or "NA" if no EC number (-> always for prodigal)
    inf2 : str
        more information found by prokka/prodigal, or "NA" if no more information
    db_xref : str
        db_xref given by Prokka ("NA" for prodigal)
    strand : str
        C (complement) or D (direct)
    start : str
        start of gene in the contig
    end : str
        end of gene in the contig
    lstopenfile : _io.TextIOWrapper
        open file where lstinfo must be written

    Returns
    -------
    str :
        lstline
    """
    locus_name = "{}.{}{}_{}".format(genome, str(cont_num).zfill(4), cont_loc,
                                     str(locus_num).zfill(5))
    # If '|' character found in those fields, replace by '_' to avoid problems while parsing
    more_info = "| {} | {} | {} | {}".format(product.replace("|", "_"),
                                        ecnum.replace("|", "_"),
                                        inf2.replace("|", "_"),
                                        db_xref.replace("|", "_"))
    lst_line = "\t".join([start, end, strand, gtype, locus_name, gene_name, more_info])
    lstopenfile.write(lst_line + "\n")
    return lst_line


def write_header(lstline, outfile):
    """
    write header to output file. Header is generated from the lst line.

    Parameters
    ----------
    lstline : str
        current line of lst file
    outfile : _io.TextIOWrapper
        open file where header must be written
    """
    start, end, _, _, name, gene_name, info = lstline.split("\t")
    size = int(end) - int(start) + 1
    towrite = " ".join([name, str(size), gene_name, info])
    outfile.write(">" + towrite + "\n")



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for annotate/annotation_functions.py
"""

import pytest
import os
import logging
import shutil
import multiprocessing

import test.test_unit.utilities_for_tests as tutil
import PanACoTA.utils as utils
from PanACoTA.annotate_module import annotation_functions as afunc


# Define variables used by several tests
DBDIR = os.path.join("test", "data", "annotate")
GEN_PATH = os.path.join(DBDIR, "genomes")
TMP_PATH = os.path.join(DBDIR, "tmp_files")
EXP_DIR = os.path.join(DBDIR, 'exp_files')
TEST_DIR = os.path.join(DBDIR, 'test_files')
GENEPATH = os.path.join(DBDIR, "generated_by_unit-tests")
LOGFILE_BASE = os.path.join(GENEPATH, "logfile.log")
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    if os.path.isdir(GENEPATH):
        content = os.listdir(GENEPATH)
        for f in content:
            assert f.startswith(".fuse")
    else:
        os.mkdir(GENEPATH)
    print("setup")

    yield
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    shutil.rmtree(GENEPATH, ignore_errors=True)
    print("teardown")


# Create a logger with sublogger inside
def my_logger(name):
    """
    logger given to function called by a subprocess
    """
    m = multiprocessing.Manager()
    q = m.Queue()
    qh = logging.handlers.QueueHandler(q)
    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    root.handlers = []
    logging.addLevelName(utils.detail_lvl(), "DETAIL")
    root.addHandler(qh)
    return q, logging.getLogger(name)


def test_check_prokka_no_outdir():
    """
    Test that prokka returns the right error message when output directory does not exist
    """
    logger = my_logger("test_check_prokka_no_outdir")
    outdir = os.path.join(GENEPATH, "outdir")
    name = "prokka_out_for_test"
    logf = "prokka.log"
    gpath = os.path.join(GENEPATH, "toto.fna")
    nbcont = 7
    assert not afunc.check_prokka(outdir, logf, name, gpath, nbcont, logger[1])
    q = logger[0]
    assert q.qsize() == 1
    msg = "Previous annotation could not run properly. Look at prokka.log for more information."
    assert q.get().message == msg


def test_check_prokka_nofna():
    """
    Check that check_prokka returns false when a tbl file is missing, and an error message
    """
    logger = my_logger("test_check_prokka_nofna")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_nofna")
    name = "prokka_out_for_test-missfna"
    gpath = "path/to/nogenome/original_name-error.fna"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + ".tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prokka.log")
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg = "prokka_out_for_test-missfna original_name-error.fna: no .fna file"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_sevfna():
    """
    Check that check_prokka returns false when there is more than 1 tbl file,
    and an error message
    """
    logger = my_logger("test_check_prokka_sevfna")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_sevfna")
    name = "prokka_out_for_test-sevfna"
    gpath = "path/to/nogenome/original_name-error.fna"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + ".tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + "2.fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prokka.log")
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg = "prokka_out_for_test-sevfna original_name-error.fna: several .fna files"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_notbl():
    """
    Check that check_prokka returns false when a tbl file is missing, and an error message
    """
    logger = my_logger("test_check_prokka_notbl")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_notbl")
    name = "prokka_out_for_test-misstbl"
    gpath = "path/to/nogenome/original_name-error.fna"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prokka.log")
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg = "prokka_out_for_test-misstbl original_name-error.fna: no .tbl file"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_sevtbl():
    """
    Check that check_prokka returns false when there is more than 1 tbl file,
    and an error message
    """
    logger = my_logger("test_check_prokka_sevtbl")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_notbl")
    name = "prokka_out_for_test-misstbl"
    gpath = "path/to/nogenome/original_name-error.fna"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + ".tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + "2.tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prokka.log")
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg = "prokka_out_for_test-misstbl original_name-error.fna: several .tbl files"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_nofaa():
    """
    Check that check_prokka returns false when a faa file is missing, and an error message
    """
    logger = my_logger("test_check_prokka_nofaa")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_nofaa")
    name = "prokka_out_for_test-missfaa"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + ".tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prokka.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg = "prokka_out_for_test-missfaa original_name.fna: no .faa file"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_sevfaa():
    """
    Check that check_prokka returns false when there is more than 1 faa file,
    and an error message
    """
    logger = my_logger("test_check_prokka_sevfaa")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_nofaa")
    name = "prokka_out_for_test-missfaa"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + ".tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + "2.faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prokka.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg = "prokka_out_for_test-missfaa original_name.fna: several .faa files"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_noffn():
    """
    Check that check_prokka returns false when a ffn file is missing, and an error message
    """
    logger = my_logger("test_check_prokka_noffn")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prokka_out_for_test-missffn"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + ".tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prokka.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg = "prokka_out_for_test-missffn original_name.fna: no .ffn file"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_sevffn():
    """
    Check that check_prokka returns false when there is more than 1 ffn file,
    and an error message
    """
    logger = my_logger("test_check_prokka_sevffn")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prokka_out_for_test-missffn"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + ".tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + "2.ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prokka.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg = "prokka_out_for_test-missffn original_name.fna: several .ffn files"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_nogff():
    """
    Check that check_prokka returns false when a ffn file is missing, and an error message
    """
    logger = my_logger("test_check_prokka_nogff")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prokka_out_for_test-missgff"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + ".tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    logf = os.path.join(GENEPATH, "prokka.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg = "prokka_out_for_test-missgff original_name.fna: no .gff file"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_sevgff():
    """
    Check that check_prokka returns false when there is more than 1 ffn file,
    and an error message
    """
    logger = my_logger("test_check_prokka_sevgff")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "out_test_noffn")
    name = "prokka_out_for_test-sevgff"
    os.makedirs(out_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".tbl"),
                    os.path.join(out_dir, name + ".tbl"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + "2.gff"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))
    logf = os.path.join(GENEPATH, "prokka.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg = "prokka_out_for_test-sevgff original_name.fna: several .gff files"
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_wrong_cont():
    """
    Check that check_prokka returns an error message when the number of contigs in tbl
    file is not as expected
    """
    logger = my_logger("test_check_prokka_wrong_cont")
    outdir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    name = "prokka_out_for_test"
    logf = os.path.join(GENEPATH, "prokka.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 10
    assert not afunc.check_prokka(outdir, logf, name, gpath, nbcont, logger[1])
    msg = ("prokka_out_for_test original_name.fna: no matching number of contigs; "
           "nbcontig=10; in tbl =6")
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg


def test_check_prokka_wrong_tbl_cds():
    """
    Check that check_prokka returns an error message when the number of CDS in tbl
    file is different from the number of headers in faa file
    """
    logger = my_logger("test_check_prokka_wrong_tbl_cds")
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    out_dir = os.path.join(GENEPATH, "res_checkProkkaWrongTbl")
    os.makedirs(out_dir)
    name = "prokka_out_for_test-wrongCDS"
    tblfile = os.path.join(TEST_DIR, name + ".tbl")
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(out_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(out_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(out_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(out_dir, name + ".gff"))

    shutil.copyfile(tblfile, os.path.join(out_dir, name + ".tbl"))
    logf = os.path.join(GENEPATH, "prokka.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 7
    assert not afunc.check_prokka(out_dir, logf, name, gpath, nbcont, logger[1])
    msg1 = ("prokka_out_for_test-wrongCDS original_name.fna: "
            "no matching number of proteins between tbl and faa; "
            "faa=14; in tbl =12")
    q = logger[0]
    assert q.qsize() == 1
    assert q.get().message == msg1


def test_check_prokka_ok():
    """
    Check that everything is ok with prokka results (tbl, faa and ffn files exist,
    and number of CDS, CRISPR and genes correspond between them)
    """
    logger = my_logger("test_check_prokka_ok")
    outdir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    name = "prokka_out_for_test"
    logf = os.path.join(GENEPATH, "prokka.log")
    gpath = "path/to/nogenome/original_name.fna"
    nbcont = 6
    assert afunc.check_prokka(outdir, logf, name, gpath, nbcont, logger[1])
    q = logger[0]
    assert q.qsize() == 0


def test_run_prokka_out_exists_ok():
    """
    Test that when the output directory already exists, and files inside are OK,
    run_prokka returns True, with a warning message indicating that prokka did not rerun.
    """
    logger = my_logger("test_run_prokka_out_exists_ok")
    utils.init_logger(LOGFILE_BASE, 0, 'prokka_out_exists_ok')
    gpath = "path/to/nogenome/original_name.fna"
    cores_prokka = 1
    name = "prokka_out_for_test"
    force = False
    nbcont = 6
    trn_file = "nofile.trn"
    arguments = (gpath, TEST_DIR, cores_prokka, name, force, nbcont, trn_file, logger[0])
    assert afunc.run_prokka(arguments)

    q = logger[0]
    assert q.qsize() == 4
    # start annotating :
    assert q.get().message.startswith("Start annotating")
    # warning prokka results folder exists:
    assert q.get().message.startswith("Prokka results folder test/data/annotate/"
                                      "test_files/"
                                      "original_name.fna-prokkaRes already exists.")
    # Results in result folder are ok
    assert q.get().message.startswith("Prokka did not run again, formatting step used already "
                                      "generated results of Prokka in "
                                      "test/data/annotate/test_files/original_name.fna-prokkaRes.")
    # End annotation:
    assert q.get().message.startswith("End annotating")


def test_run_prokka_out_exists_error():
    """
    Test that when the output directory already exists, and 1 file is missing,
    run_prokka returns False, and writes the warning message saying that prokka did not
    rerun, + the warning message for the missing file(s).
    """
    logger = my_logger("test_run_prokka_out_exists_error")
    utils.init_logger(LOGFILE_BASE, 0, 'prokka_out_error')
    ori_prok_dir = os.path.join(TEST_DIR, "original_name.fna-prokkaRes")
    ori_name = "prokka_out_for_test"
    new_prok_dir = os.path.join(GENEPATH, "original_name-error-prokkaRes")
    name = "prokka_out_for_test-wrongCDS"
    os.makedirs(new_prok_dir)
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".fna"),
                    os.path.join(new_prok_dir, name + ".fna"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".ffn"),
                    os.path.join(new_prok_dir, name + ".ffn"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".faa"),
                    os.path.join(new_prok_dir, name + ".faa"))
    shutil.copyfile(os.path.join(ori_prok_dir, ori_name + ".gff"),
                    os.path.join(new_prok_dir, name + ".gff"))
    gpath = "path/to/nogenome/original_name-error"
    cores_prokka = 1
    force = False
    nbcont = 6
    trn_file = "nofile.trn"
    arguments = (gpath, GENEPATH, cores_prokka, name, force, nbcont, trn_file, logger[0])
    assert not afunc.run_prokka(arguments)
    q = logger[0]
    assert q.qsize() == 4
    # start annotating :
    assert q.get().message.startswith("Start annotating")
    # warning prokka results folder exists:
    assert q.get().message == ("Prokka results folder test/data/annotate/generated_by_unit-tests/"
                               "original_name-error-prokkaRes already exists.")
    # error, no tbl file
    assert q.get().message == "prokka_out_for_test-wrongCDS original_name-error: no .tbl file"
    # warning, files in outdir are not as expected
    assert q.get().message.startswith("Problems in the files contained in your already existing "
                                      "output dir (test/data/annotate/generated_by_unit-tests/"
                                      "original_name-error-prokkaRes)")


def test_run_prokka_out_exists_force():
    """
    Test that when the output directory already exists with wrong files, but force is on,
    prokka is rerun and outputs the right files
    """
    logger = my_logger("test_run_prokka_out_exists_force")
    utils.init_logger(LOGFILE_BASE, 0, 'force')
    gpath = os.path.join(GEN_PATH, "H299_H561.fasta")
    out_prokdir = os.path.join(GENEPATH, "H299_H561.fasta-prokkaRes")
    name = "test_runprokka_H299"
    # Put empty tbl, faa, ffn files in prokka output dir, to check that they are overridden
    os.makedirs(out_prokdir)
    open(os.path.join(out_prokdir, name + ".tbl"), "w").close()
    open(os.path.join(out_prokdir, name + ".faa"), "w").close()
    open(os.path.join(out_prokdir, name + ".ffn"), "w").close()
    cores_prokka = 2
    force = True
    nbcont = 3
    trn_file = "nofile.trn"
    arguments = (gpath, GENEPATH, cores_prokka, name, force, nbcont, trn_file, logger[0])
    assert afunc.run_prokka(arguments)
    # As we used 'force', tbl, faa and ffn files, which were empty, must have been replaced
    # by the prokka output
    exp_dir = os.path.join(EXP_DIR, "H299_H561.fasta-short-contig.fna-prokkaRes",
                           "test_runprokka_H299")
    out_tbl = os.path.join(out_prokdir, name + ".tbl")
    out_faa = os.path.join(out_prokdir, name + ".faa")
    out_ffn = os.path.join(out_prokdir, name + ".ffn")
    assert os.path.isfile(out_tbl)
    # For tbl file, check that, at least, the 3 contigs were considered,
    # and that the number of CDS is as expected.
    # Before, we checked that the output
    # was exactly as expected. But it changes with the different versions of prokka, so
    # we cannot compare the whole file.
    with open(out_tbl, "r") as outt:
        lines = [line.strip() for line in outt.readlines()]
        # Check that there are 3 contigs
        feature = 0
        for line in lines:
            if 'Feature' in line:
                feature += 1
        assert feature == 3
        # Check that there are 16 CDS
        CDS = 0
        for line in lines:
            if "CDS" in line:
                CDS += 1
        assert CDS == 16
    # Check that faa and ffn files are as expected
    assert os.path.isfile(out_faa)
    assert tutil.compare_order_content(exp_dir + ".faa", out_faa)
    assert os.path.isfile(out_ffn)
    assert tutil.compare_order_content(exp_dir + ".ffn", out_ffn)
    q = logger[0]
    assert q.qsize() == 4
    assert q.get() .message.startswith("Start annotating test_runprokka_H299 from test/data/"
                                       "annotate/genomes/H299_H561.fasta with Prokka")
    assert q.get() .message == ("Prokka results folder already exists, but removed because "
                                "--force option used")
    assert q.get().message == ("Prokka command: prokka "
                               "--outdir test/data/annotate/generated_by_unit-tests/"
                               "H299_H561.fasta-prokkaRes --cpus 2 --prefix test_runprokka_H299 "
                               "--centre prokka test/data/annotate/genomes/H299_H561.fasta")
    assert q.get() .message.startswith("End annotating test_runprokka_H299 "
                                       "from test/data/annotate/genomes/H299_H561.fasta")


def test_run_prokka_out_doesnt_exist_ok():
    """
    Test that when the output directory does not exist, it creates it, and runs prokka
    with all expected outfiles
    """
    logger = my_logger("test_run_prokka_out_doesnt_exist")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_prokka_out_doesnt_exist')
    gpath = os.path.join(GEN_PATH, "H299_H561.fasta")
    out_dir = os.path.join(GENEPATH, "H299_H561.fasta-prokkaRes")
    cores_prokka = 2
    name = "test_runprokka_H299"
    force = False
    nbcont = 3
    trn_file = "nofile.trn"
    arguments = (gpath, GENEPATH, cores_prokka, name, force, nbcont, trn_file, logger[0])
    assert afunc.run_prokka(arguments)
    # Check content of tbl, ffn and faa files
    exp_dir = os.path.join(EXP_DIR, "H299_H561.fasta-short-contig.fna-prokkaRes",
                           "test_runprokka_H299")
    out_tbl = os.path.join(out_dir, name + ".tbl")
    out_faa = os.path.join(out_dir, name + ".faa")
    out_ffn = os.path.join(out_dir, name + ".ffn")
    out_gff = os.path.join(out_dir, name + ".gff")
    assert os.path.isfile(out_tbl)
    # For tbl file, check that, at least, the 3 contigs were considered,
    # and that the number of CDS is as expected.
    # Before, we checked that the output
    # was exactly as expected. But it changes with the different versions of prokka, so
    # we cannot compare the whole file.
    with open(out_tbl, "r") as outt:
        lines = [line.strip() for line in outt.readlines()]
        # Check that there are 3 contigs
        feature = 0
        for line in lines:
            if 'Feature' in line:
                feature += 1
        assert feature == 3
        # Check that there are 16 CDS
        CDS = 0
        for line in lines:
            if "CDS" in line:
                CDS += 1
        assert CDS == 16
    # Check that faa and ffn files are as expected
    assert os.path.isfile(out_faa)
    assert tutil.compare_order_content(exp_dir + ".faa", out_faa)
    assert os.path.isfile(out_ffn)
    assert tutil.compare_order_content(exp_dir + ".ffn", out_ffn)
    q = logger[0]
    assert q.qsize() == 3
    assert q.get().message.startswith("Start annotating")
    assert q.get().message == ("Prokka command: prokka "
                               "--outdir test/data/annotate/generated_by_unit-tests/"
                               "H299_H561.fasta-prokkaRes --cpus 2 --prefix test_runprokka_H299 "
                               "--centre prokka test/data/annotate/genomes/H299_H561.fasta")
    assert q.get().message.startswith("End annotating")


def test_run_prokka_out_problem_running():
    """
    Check that when a problem occurs while trying to run prokka, run_prokka returns False,
    and the error message indicating to read in the log why it couldn't run
    """
    logger = my_logger("test_run_prokka_out_problem_running")
    utils.init_logger(LOGFILE_BASE, 0, 'test_run_prokka_out_problem_running')
    gpath = os.path.join(GEN_PATH, "H299_H561bis.fasta")
    cores_prokka = 2
    name = "test_runprokka_H299-error"
    force = False
    nbcont = 3
    logf = os.path.join(GENEPATH, "H299_H561.fasta-prokka.log")
    trn_file = "nofile.trn"
    arguments = (gpath, GENEPATH, cores_prokka, name, force, nbcont, trn_file, logger[0])
    assert not afunc.run_prokka(arguments)
    q = logger[0]
    assert q.qsize() == 3
    assert q.get().message.startswith("Start annotating")
    assert q.get().message == ("Prokka command: prokka "
                               "--outdir test/data/annotate/generated_by_unit-tests/"
                               "H299_H561bis.fasta-prokkaRes --cpus 2 "
                               "--prefix test_runprokka_H299-error "
                               "--centre prokka test/data/annotate/genomes/H299_H561bis.fasta")
    assert q.get().message == ("Error while trying to run prokka on test_runprokka_H299-error "
                               "from test/data/annotate/genomes/H299_H561bis.fasta")



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the fasttree_func submodule in tree module
"""
import subprocess
import os
import logging
import pytest
import shutil

import PanACoTA.tree_module.fasttree_func as ft

# Define common variables
from PanACoTA import utils
from . import utilities as tutil


ALPATH = os.path.join("test", "data", "align")
ALIGNMENT = os.path.join(ALPATH, "exp_files", "exp_pers4genomes.grp.aln")
TREEPATH = os.path.join("test", "data", "tree")
GENEPATH = os.path.join(TREEPATH, "generated_by_unit-tests")
LOGFILE_BASE = "log_test_fasttree"
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    utils.init_logger(LOGFILE_BASE, 0, 'test_fasttree', verbose=1)
    os.mkdir(GENEPATH)
    print("setup")

    yield
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    shutil.rmtree(GENEPATH)
    print("teardown")


def test_def_nb_threads():
    """
    Test that when changing the number of threads, it is taken into account as expected by
    fasttree binary.
    """
    thread = 10
    ft.define_nb_threads(thread)
    res = subprocess.Popen("FastTreeMP", stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout, _ = res.communicate()
    assert "OpenMP (10 threads)" in stdout.decode()
    # Put back 1 for next tests, and check that it is taken into account
    ft.define_nb_threads(1)
    res1 = subprocess.Popen("FastTreeMP", stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout1, _ = res1.communicate()
    assert "OpenMP (1 threads)" in stdout1.decode()


def test_run_fasttree_default(caplog):
    """
    Test that when running fasttree without bootstrap, and with default model, it returns a file
    in the expected format (all branches have lengths, no bootstrap value).
    """
    caplog.set_level(logging.DEBUG)
    boot = None
    model = "-gtr"
    quiet = False
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.fasttree_tree.nwk")
    ft.run_fasttree(ALIGNMENT, boot, GENEPATH, model, quiet)
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.fasttree.log")
    assert os.path.isfile(logs)
    assert "Running FasttreeMP..." in caplog.text
    assert ("FastTreeMP -nt -gtr -noml -nocat -nosupport "
            "-log test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.fasttree.log "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln")in caplog.text
    assert tutil.is_tree_lengths(treefile)
    assert not tutil.is_tree_bootstrap(treefile)


def test_run_fasttree_boot(caplog):
    """
    Test that when running fasttree with 10 bootstrap, and with default model, it returns a file
    in the expected format (all branches have lengths, and bootstrap value).
    """
    caplog.set_level(logging.DEBUG)
    boot = 10
    model = "-gtr"
    quiet = False
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.fasttree_tree.nwk")
    ft.run_fasttree(ALIGNMENT, boot, GENEPATH, model, quiet)
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.fasttree.log")
    assert os.path.isfile(logs)
    assert "Running FasttreeMP..." in caplog.text
    assert ("FastTreeMP -nt -gtr -noml -nocat -boot 10 "
            "-log test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.fasttree.log "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln") in caplog.text
    assert not tutil.is_tree_lengths(treefile)
    assert tutil.is_tree_bootstrap(treefile)


def test_run_fasttree_boot_noname_jc(caplog):
    """
    Test that when running fasttree with bootstrap, and with JC model, it returns a file
    in the expected format (all branches have lengths, and bootstrap value).
    """
    caplog.set_level(logging.DEBUG)
    boot = 100
    model = ""
    quiet = False
    ft.run_fasttree(ALIGNMENT, boot, GENEPATH, model, quiet)
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.fasttree_tree.nwk")
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.fasttree.log")
    assert os.path.isfile(logs)
    assert "Running FasttreeMP..." in caplog.text
    assert ("FastTreeMP -nt  -noml -nocat -boot 100 "
            "-log test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.fasttree.log "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln") in caplog.text
    assert not tutil.is_tree_lengths(treefile)
    assert tutil.is_tree_bootstrap(treefile)


def test_run_fasttree_default_quiet(caplog):
    """
    Test that when running fasttree without bootstrap, and with default model, it returns a file
    in the expected format (all branches have lengths, no bootstrap value).
    Check that no bug when launch in quiet mode
    """
    caplog.set_level(logging.DEBUG)
    boot = None
    model = "-gtr"
    quiet = True
    ft.run_fasttree(ALIGNMENT, boot, GENEPATH, model, quiet)
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.fasttree_tree.nwk")
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.fasttree.log")
    assert os.path.isfile(logs)
    assert "Running FasttreeMP..." in caplog.text
    assert ("FastTreeMP -nt -gtr -noml -nocat -nosupport "
            "-log test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.fasttree.log "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln") in caplog.text
    assert tutil.is_tree_lengths(treefile)
    assert not tutil.is_tree_bootstrap(treefile)


def test_run_tree(caplog):
    """
    Test that when running run_tree it defines the expected number of threads, and outputs the
    expected tree.
    """
    # without bootstrap
    caplog.set_level(logging.DEBUG)
    boot = None
    quiet = False
    threads = 15
    model = "-gtr"
    ft.run_tree(ALIGNMENT, boot, ".", quiet, threads, model=model)
    # Check it used 15 threads
    res = subprocess.Popen("FastTreeMP", stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout, _ = res.communicate()
    assert "OpenMP (15 threads)" in stdout.decode()
    # Check outputs
    out_tree = "exp_pers4genomes.grp.aln.fasttree_tree.nwk"
    assert os.path.isfile(out_tree)
    out_log = "exp_pers4genomes.grp.aln.fasttree.log"
    assert os.path.isfile(out_log)
    assert "Running FasttreeMP..." in caplog.text
    assert ("FastTreeMP -nt -gtr -noml -nocat -nosupport "
            "-log ./exp_pers4genomes.grp.aln.fasttree.log "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln") in caplog.text
    assert tutil.is_tree_lengths(out_tree)
    assert not tutil.is_tree_bootstrap(out_tree)
    os.remove(out_tree)
    os.remove(out_log)

    # with bootstrap
    boot = 127
    quiet = True
    ft.run_tree(ALIGNMENT, boot, ".", quiet, 1, model=model)
    res = subprocess.Popen("FastTreeMP", stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout, _ = res.communicate()
    assert "OpenMP (1 threads)" in stdout.decode()
    out_tree = "exp_pers4genomes.grp.aln.fasttree_tree.nwk"
    assert os.path.isfile(out_tree)
    out_log = "exp_pers4genomes.grp.aln.fasttree.log"
    assert os.path.isfile(out_log)
    assert "Running FasttreeMP..." in caplog.text
    assert ("FastTreeMP -nt -gtr -noml -nocat -boot 127 "
            "-log ./exp_pers4genomes.grp.aln.fasttree.log "
            "test/data/align/exp_files/exp_pers4genomes.grp.aln") in caplog.text
    assert not tutil.is_tree_lengths(out_tree)
    assert tutil.is_tree_bootstrap(out_tree)
    os.remove(out_tree)
    os.remove(out_log)



#!/usr/bin/env python3
# coding: utf-8

"""
Unit tests for the iqtree_func submodule in tree module
"""
import subprocess
import os
import logging
import pytest
import shutil

import PanACoTA.tree_module.iqtree_func as ft

# Define common variables
from PanACoTA import utils
from . import utilities as tutil


ALPATH = os.path.join("test", "data", "align")
ALIGNMENT = os.path.join(ALPATH, "exp_files", "exp_pers4genomes.grp.aln")
TREEPATH = os.path.join("test", "data", "tree")
GENEPATH = os.path.join(TREEPATH, "generated_by_unit-tests")
LOGFILE_BASE = "log_test_iqtree"
LOGFILES = [LOGFILE_BASE + ext for ext in [".log", ".log.debug", ".log.details", ".log.err"]]


@pytest.fixture(autouse=True)
def setup_teardown_module():
    """
    Remove log files at the end of this test module

    Before each test:
    - init logger
    - create directory to put generated files

    After:
    - remove all log files
    - remove directory with generated results
    """
    utils.init_logger(LOGFILE_BASE, 0, 'test_iqtree', verbose=1)
    os.mkdir(GENEPATH)
    print("setup")

    yield
    for f in LOGFILES:
        if os.path.exists(f):
            os.remove(f)
    shutil.rmtree(GENEPATH)
    print("teardown")


def test_run_iqtree_default(caplog):
    """
    Test that when running iqtree without bootstrap, and with default model, it returns a file
    in the expected format (all branches have lengths, no bootstrap value, treefile
    created with expected name).
    """
    caplog.set_level(logging.DEBUG)
    boot = None
    threads = 1
    quiet = False
    model = "GTR"
    fast = False
    ft.run_tree(ALIGNMENT, boot, GENEPATH, quiet, threads, model=model, wb="", mem="",
                s="iqtree", f=fast)
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.treefile")
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.log")
    assert os.path.isfile(logs)
    assert "Running IQtree..." in caplog.text
    assert ("iqtree -s test/data/align/exp_files/exp_pers4genomes.grp.aln "
            "-nt 1 -m GTR    -st DNA "
            "-pre test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.iqtree_tree "
            "-quiet") in caplog.text
    assert tutil.is_tree_lengths(treefile)
    assert not tutil.is_tree_bootstrap(treefile)


def test_run_iqtree2_default(caplog):
    """
    Test that when running iqtree2 without bootstrap, and with default model, it returns a file
    in the expected format (all branches have lengths, no bootstrap value, treefile
    created with expected name).
    """
    caplog.set_level(logging.DEBUG)
    boot = None
    threads = 1
    quiet = False
    model = "GTR"
    fast = False
    ft.run_tree(ALIGNMENT, boot, GENEPATH, quiet, threads, model=model, wb="", mem="",
                s="iqtree2", f=fast)
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.treefile")
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.log")
    assert os.path.isfile(logs)
    assert "Running IQtree..." in caplog.text
    assert ("iqtree2 -s test/data/align/exp_files/exp_pers4genomes.grp.aln "
            "-T 1 -m GTR    --seqtype DNA "
            "--prefix test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.iqtree_tree "
            "--quiet") in caplog.text
    assert tutil.is_tree_lengths(treefile)
    assert not tutil.is_tree_bootstrap(treefile)


def test_run_iqtree_boot_quiet_TVM(caplog):
    """
    Test that when running iqtree with bootstrap, and with TVM model, it returns a file
    in the expected format (all branches have lengths and bootstrap value).
    Treefile keeps the name given as input.
    """
    caplog.set_level(logging.DEBUG)
    boot = 1000
    threads = 1
    quiet = True
    model = "TVM"
    fast = False
    ft.run_tree(ALIGNMENT, boot, GENEPATH, quiet, threads, model=model, wb="", mem="",
                f=fast, s="iqtree")
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.treefile")
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.log")
    assert os.path.isfile(logs)
    assert "Running IQtree..." in caplog.text
    assert ("iqtree -s test/data/align/exp_files/exp_pers4genomes.grp.aln "
            "-nt 1 -m TVM  -bb 1000  -st DNA "
            "-pre test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.iqtree_tree") in caplog.text
    assert not tutil.is_tree_lengths(treefile)
    assert tutil.is_tree_bootstrap(treefile)


def test_run_iqtree_boot_write_boot(caplog):
    """
    Test that when running iqtree with bootstrap, writing bootstrap trees,
    and with default model, it returns a file
    in the expected format (all branches have lengths and bootstrap value).
    Treefile keeps the name given as input.
    """
    caplog.set_level(logging.DEBUG)
    boot = 1000
    threads = 1
    quiet = True
    model = "GTR"
    fast = False
    ft.run_tree(ALIGNMENT, boot, GENEPATH, quiet, threads, model=model, wb=True, mem="",
                f=fast, s="iqtree")
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.treefile")
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.log")
    assert os.path.isfile(logs)
    assert "Running IQtree..." in caplog.text
    assert ("iqtree -s test/data/align/exp_files/exp_pers4genomes.grp.aln "
            "-nt 1 -m GTR  -bb 1000 -wbt -st DNA "
            "-pre test/data/tree/generated_by_unit-tests/"
            "exp_pers4genomes.grp.aln.iqtree_tree") in caplog.text
    assert not tutil.is_tree_lengths(treefile)
    assert tutil.is_tree_bootstrap(treefile)


def test_run_iqtree2_boot_write_boot(caplog):
    """
    Test that when running iqtree2 with bootstrap, writing bootstrap trees,
    and with default model, it returns a file
    in the expected format (all branches have lengths and bootstrap value).
    Treefile is named as expected.
    """
    caplog.set_level(logging.DEBUG)
    boot = 1000
    threads = 1
    quiet = False
    model = "GTR"
    treefile = ""
    fast = False
    ft.run_tree(ALIGNMENT, boot, GENEPATH, quiet, threads, model=model, wb=True, mem="",
                f=fast, s="iqtree2")
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.treefile")
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.log")
    assert os.path.isfile(logs)
    assert "Running IQtree..." in caplog.text
    assert ("iqtree2 -s test/data/align/exp_files/exp_pers4genomes.grp.aln "
            "-T 1 -m GTR  -B 1000 --boot-trees --seqtype DNA "
            "--prefix test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.iqtree_tree "
            "--quiet") in caplog.text
    assert not tutil.is_tree_lengths(treefile)
    assert tutil.is_tree_bootstrap(treefile)


def test_run_iqtree_fast_mem(caplog):
    """
    Test that when running iqtree without bootstrap, with F81 model,
    with fast option and giving a memory limit, it returns a file
    in the expected format (all branches have lengths, no bootstrap value).
    Treefile is named as expected
    """
    caplog.set_level(logging.DEBUG)
    boot = None
    threads = 1
    quiet = False
    model = "F81"
    fast = True
    ft.run_tree(ALIGNMENT, boot, GENEPATH, quiet, threads, model=model, wb="", mem="4GB",
                s="iqtree", f=fast)
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.treefile")
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.log")
    assert os.path.isfile(logs)
    assert "Running IQtree..." in caplog.text
    assert ("iqtree -s test/data/align/exp_files/exp_pers4genomes.grp.aln "
            "-nt 1 -m F81 -mem 4GB   -st DNA "
            "-pre test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.iqtree_tree "
            "-quiet -fast") in caplog.text
    assert tutil.is_tree_lengths(treefile)
    assert not tutil.is_tree_bootstrap(treefile)


def test_run_iqtree2_fast_mem_quiet(caplog):
    """
    Test that when running iqtree without bootstrap, with default model,
    with fast option and giving a memory limit, it returns a file
    in the expected format (all branches have lengths, no bootstrap value).
    Treefile is named as expected
    """
    caplog.set_level(logging.DEBUG)
    boot = None
    threads = 1
    quiet = True
    model = "GTR"
    fast = True
    ft.run_tree(ALIGNMENT, boot, GENEPATH, quiet, threads, model=model, wb="", mem="4GB",
                s="iqtree2", f=fast)
    treefile = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.treefile")
    assert os.path.isfile(treefile)
    logs = os.path.join(GENEPATH, "exp_pers4genomes.grp.aln.iqtree_tree.log")
    assert os.path.isfile(logs)
    assert "Running IQtree..." in caplog.text
    assert ("iqtree2 -s test/data/align/exp_files/exp_pers4genomes.grp.aln -T 1 "
            "-m GTR --mem 4GB   --seqtype DNA "
            "--prefix test/data/tree/generated_by_unit-tests/exp_pers4genomes.grp.aln.iqtree_tree "
            "--quiet -fast") in caplog.text
    assert tutil.is_tree_lengths(treefile)
    assert not tutil.is_tree_bootstrap(treefile)




#!/usr/bin/env python3
# coding: utf-8

"""
Tests for make script, installing PanACoTA according to already existing dependencies
Here, all dependencies are installed
"""
import os
import glob
import pytest
import subprocess
import shlex

from . import utilities as utils


@pytest.fixture
def install_panacota():
    cmd = "python3 make"
    error = "Error installing"
    utils.run_cmd(cmd, error)


def teardown_function(function):
    """
    Uninstall PanACoTA and installed dependencies at the end of each test
    """
    print("TEARDOWN\n")
    cmd = "python3 make uninstall"
    error = "Error uninstall"
    utils.run_cmd(cmd, error)
    os.remove("install.log")
    print("cleaning repo")


def test_install_panacota():
    """
    Test that when installing from a computer containing only all dependencies, it returns a message without any warning: everything is ok
    """
    cmd = "python3 make"
    error = "Error trying to install PanACoTA from base"
    assert not utils.check_installed("PanACoTA")
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("mash")
    assert utils.check_installed("FastTreeMP")
    # Install panacota
    utils.run_cmd(cmd, error)
    assert utils.check_installed("PanACoTA")
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("mash")
    assert utils.check_installed("FastTreeMP")
    # Check that panacota is installed (pip3 module exists)
    assert utils.is_package_installed("PanACoTA")
    # Check that it is installed in "final mode"
    cmd = "pip3 show PanACoTA"
    err = "error pip3"
    stdout = "stdout_pip3show.out"
    with open(stdout, "w") as stdof:
        utils.run_cmd(cmd, err, stdout=stdof, stderr=stdof)
    with open(stdout, "r") as stdof:
        for line in stdof:
            if line.startswith("Location"):
                loc = line.split()[-1]
                assert glob.glob(os.path.join(loc, r'PanACoTA*dist-info'))
    os.remove(stdout)
    # Check not installed in user mode
    cmd = "pip list --user"
    list_user_packages = str(subprocess.check_output(shlex.split(cmd)))
    assert "PanACoTA" not in list_user_packages
    # Check output logfile content. Check that all content is present, in any order.
    logfile = "install.log"
    content = ["Installing PanACoTA...", "DONE"]
    with open(logfile, "r") as logf:
        logf_content = "".join(logf.readlines())
        for linec in content:
            assert linec in logf_content
    # Check that needed packages are installed
    assert utils.is_package_installed("argparse")
    assert utils.is_package_installed("progressbar")
    assert utils.is_package_installed("numpy")
    assert utils.is_package_installed("matplotlib")
    assert utils.is_package_installed("Bio")
    print("test_installed done")


def test_upgrade(install_panacota):
    """
    Test upgrading PanACoTA when dependencies are still installed
    """
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("FastTreeMP")
    assert utils.check_installed("PanACoTA")
    # Upgrade PanACoTA
    cmd = "python3 make upgrade"
    error = "Error upgrade"
    utils.run_cmd(cmd, error)
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("PanACoTA")
    assert utils.check_installed("FastTreeMP")
    logfile = "install.log"
    with open(logfile, "r") as logf:
        lines = logf.readlines()
        assert len(lines) == 2
        assert "Upgrading PanACoTA" in lines[0]
        assert "DONE" in lines[1]
    print("test_upgrade done")


def test_upgrade_notinstalled():
    """
    Test upgrading PanACoTA when dependencies are not installed (only barrnap),
    and PanACoTA is not installed. It just installs PanACoTA, without prokka dep
    """
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("FastTreeMP")
    assert not utils.check_installed("PanACoTA")
    cmd = "python3 make upgrade"
    error = "Error upgrade"
    utils.run_cmd(cmd, error)
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("PanACoTA")
    assert utils.check_installed("FastTreeMP")
    logfile = "install.log"
    with open(logfile, "r") as logf:
        lines = logf.readlines()
        assert len(lines) == 2
        assert "Upgrading PanACoTA" in lines[0]
        assert "DONE" in lines[1]
    os.remove(logfile)
    print("test_upgrade_notinstalled done")


def test_uninstall(install_panacota):
    """
    Test uninstalling PanACoTA when dependencies are still installed
    """
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("PanACoTA")
    assert utils.check_installed("FastTreeMP")
    cmd = "python3 make uninstall"
    error = "Error uninstalling"
    utils.run_cmd(cmd, error)
    assert not utils.check_installed("PanACoTA")
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("mash")
    assert utils.check_installed("FastTreeMP")
    assert utils.check_installed("quicktree")
    assert utils.check_installed("iqtree") or utils.check_installed("iqtree2")
    logfile = "install.log"
    with open(logfile, "r") as logf:
        lines = logf.readlines()
        assert len(lines) == 2
        assert "Uninstalling PanACoTA" in lines[0]
        assert "DONE" in lines[1]
    print("test_uninstall done")


def test_develop():
    """
    Test installing PanACoTA in developer mode, when prokka and barrnap are already installed
    """
    assert not utils.check_installed("PanACoTA")
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    cmd = "python3 make develop"
    error = "Error develop"
    utils.run_cmd(cmd, error)
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("PanACoTA")
    # Check installed in developper mode (egg-info file is present)
    cmd = "pip3 show PanACoTA"
    err = "error pip3"
    stdout = "stdout_pip3show.out"
    with open(stdout, "w") as stdof:
        utils.run_cmd(cmd, err, stdout=stdof, stderr=stdof)
    # Check installation in develop mode
    with open(stdout, "r") as stdof:
        lines = stdof.readlines()
        for line in lines:
            if line.startswith("Location"):
                loc = line.split()[-1]
                assert glob.glob(os.path.join(loc, r'PanACoTA*egg-info'))
    os.remove(stdout)
    # Check not installed in user mode
    cmd = "pip list --user"
    list_user_packages = str(subprocess.check_output(shlex.split(cmd)))
    assert "PanACoTA" not in list_user_packages
    # check logfile content
    logfile = "install.log"
    content = ["Installing PanACoTA...",
               "Installing developer packages needed for PanACoTA", "DONE"]
    with open(logfile, "r") as logf:
        logf_content = "".join(logf.readlines())
        for linec in content:
            assert linec in logf_content
    # Check that needed packages are installed
    assert utils.is_package_installed("argparse")
    assert utils.is_package_installed("progressbar")
    assert utils.is_package_installed("numpy")
    assert utils.is_package_installed("matplotlib")
    assert utils.is_package_installed("Bio")
    assert utils.is_package_installed("sphinx")
    assert utils.is_package_installed("numpydoc")
    assert utils.is_package_installed("pytest")
    assert utils.is_package_installed("coverage")
    print("test_develop done")


def test_install_user():
    """
    Test that when installing from a computer in user mode, it really installs
    PanACoTA in user mode (pip showing PanACoTA when asking for user installed packages)
    """
    cmd = "python3 make --user"
    error = "Error trying to install PanACoTA from base"
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert not utils.check_installed("PanACoTA")
    assert utils.check_installed("FastTreeMP")
    utils.run_cmd(cmd, error)
    assert utils.check_installed("barrnap")
    assert utils.check_installed("prokka")
    assert utils.check_installed("PanACoTA")
    assert utils.check_installed("FastTreeMP")
    # Check logfile content
    logfile = "install.log"
    content = ["Installing PanACoTA in user mode...", "DONE"]
    with open(logfile, "r") as logf:
        logf_content = "".join(logf.readlines())
        for linec in content:
            assert linec in logf_content
    # Check that needed packages are installed
    assert utils.is_package_installed("argparse")
    assert utils.is_package_installed("progressbar")
    assert utils.is_package_installed("numpy")
    assert utils.is_package_installed("matplotlib")
    assert utils.is_package_installed("Bio")
